# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from typing import Any, Dict, Union


def keys_to_lower(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    """Returns new dictionary content of given `dictionary` but its `keys` in lower case"""
    return {keys.lower(): value for keys, value in dictionary.items()}


def ensure_is_list(value: Any) -> list:
    """Returns a list: Either the provided `value` if it is a list, or a new list containing the provided value"""
    if isinstance(value, list):
        return value
    else:
        return [value]


def ensure_path_exists(path: Union[Path, str]):
    """Creates a specified path if not already existent"""
    Path(path).mkdir(parents=True, exist_ok=True)


def clean_up_file_name(name: str) -> str:
    """Returns given `name` with replacements defined in `replace_map`"""
    replace_map = {" ": "_", ":": "_", "/": "-"}
    return name.translate(str.maketrans(replace_map))



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import ast
from typing import List, Dict, Any, Optional, Tuple

from fameprotobuf.DataStorage_pb2 import DataStorage
from fameprotobuf.Field_pb2 import NestedField
from fameprotobuf.InputFile_pb2 import InputData

from fameio.source.logs import log
from fameio.source.scenario import GeneralProperties, Agent, Contract, Scenario
from fameio.source.schema import Schema, AttributeSpecs, AttributeType
from fameio.source.series import TimeSeriesManager


class InputConversionException(Exception):
    """An Exception indication an error during reconstruction of input from its protobuf representation"""

    pass


class InputDao:
    """Data access object for inputs saved in protobuf"""

    _ERR_NO_INPUTS = "No input data found on file."
    _ERR_MULTIPLE_INPUTS = "File corrupt. More than one input section found on file."

    _FIELD_NAME_MAP: Dict = {
        AttributeType.STRING: "stringValue",
        AttributeType.ENUM: "stringValue",
        AttributeType.INTEGER: "intValue",
        AttributeType.DOUBLE: "doubleValue",
        AttributeType.LONG: "longValue",
        AttributeType.TIME_SERIES: "seriesId",
        AttributeType.BLOCK: "field",
    }

    def __init__(self) -> None:
        self._inputs: List[InputData] = []
        self._timeseries_manager: TimeSeriesManager = TimeSeriesManager()
        self._schema: Optional[Schema] = None

    def store_inputs(self, data_storages: List[DataStorage]) -> None:
        """
        Extracts and stores Inputs in given DataStorages - if such are present

        Args:
            data_storages: to be scanned for InputData
        """
        self._inputs.extend([data_storage.input for data_storage in data_storages if data_storage.HasField("input")])

    def recover_inputs(self) -> Tuple[TimeSeriesManager, Scenario]:
        """
        Recovers inputs to GeneralProperties, Schema, Agents, Contracts, Timeseries

        Return:
            recovered timeseries and scenario

        Raises:
            InputConversionException: if inputs could not be recovered
        """
        input_data = self._get_input_data()
        self._schema = self._get_schema(input_data)
        scenario = Scenario(self._schema, self._get_general_properties(input_data))
        for contract in self._get_contracts(input_data):
            scenario.add_contract(contract)

        self._init_timeseries(input_data)
        for agent in self._get_agents(input_data):
            scenario.add_agent(agent)

        return self._timeseries_manager, scenario

    def _get_input_data(self) -> InputData:
        """
        Check that exactly one previously extracted input data exist, otherwise raises an exception

        Raises:
            InputConversionException: if no or more than one input is present
        """
        if not self._inputs:
            log().error(self._ERR_NO_INPUTS)
            raise InputConversionException(self._ERR_NO_INPUTS)
        if len(self._inputs) > 1:
            log().error(self._ERR_MULTIPLE_INPUTS)
            raise InputConversionException(self._ERR_MULTIPLE_INPUTS)
        return self._inputs[0]

    @staticmethod
    def _get_schema(input_data: InputData) -> Schema:
        """Read and return Schema from given `input_data`"""
        return Schema.from_string(input_data.schema)

    @staticmethod
    def _get_general_properties(input_data: InputData) -> GeneralProperties:
        """Read and return GeneralProperties from given `input_data`"""
        return GeneralProperties(
            run_id=input_data.runId,
            simulation_start_time=input_data.simulation.startTime,
            simulation_stop_time=input_data.simulation.stopTime,
            simulation_random_seed=input_data.simulation.randomSeed,
            output_process=input_data.output.process,
            output_interval=input_data.output.interval,
        )

    @staticmethod
    def _get_contracts(input_data: InputData) -> List[Contract]:
        """Read and return Contracts from given `input_data`"""
        return [
            Contract(
                sender_id=contract.senderId,
                receiver_id=contract.receiverId,
                product_name=contract.productName,
                delivery_interval=contract.deliveryIntervalInSteps,
                first_delivery_time=contract.firstDeliveryTime,
                expiration_time=contract.expirationTime,
                meta_data=ast.literal_eval(contract.metadata),
            )
            for contract in input_data.contract
        ]

    def _init_timeseries(self, input_data: InputData) -> None:
        """Read timeseries from given `input_data` and initialise TimeSeriesManager"""
        self._timeseries_manager.reconstruct_time_series(list(input_data.timeSeries))

    def _get_agents(self, input_data: InputData) -> List[Agent]:
        """Read and return Agents from given `input_data`"""
        agents = []
        for agent_dao in input_data.agent:
            agent = Agent(
                agent_id=agent_dao.id, type_name=agent_dao.className, meta_data=ast.literal_eval(agent_dao.metadata)
            )
            attribute_dict = self._get_attributes(
                list(agent_dao.field), self._schema.agent_types[agent_dao.className].attributes
            )
            agent.init_attributes_from_dict(attribute_dict)
            agents.append(agent)
        return agents

    def _get_attributes(self, fields: List[NestedField], schematics: Dict[str, AttributeSpecs]) -> Dict[str, Any]:
        """Read and return Attributes as Dictionary from given list of fields"""
        attributes: Dict[str, Any] = {}
        for field in fields:
            attributes[field.fieldName] = self._get_field_value(field, schematics[field.fieldName])
        return attributes

    def _get_field_value(self, field: NestedField, schematic: AttributeSpecs) -> Any:
        """Extracts and returns value(s) of given `field`"""
        attribute_type: AttributeType = schematic.attr_type
        value = field.__getattribute__(self._FIELD_NAME_MAP[attribute_type])
        if attribute_type is AttributeType.TIME_SERIES:
            return self._timeseries_manager.get_reconstructed_series_by_id(field.seriesId)
        elif attribute_type is AttributeType.BLOCK:
            if schematic.is_list:
                return [self._get_attributes(list(entry.field), schematic.nested_attributes) for entry in field.field]
            else:
                return self._get_attributes(list(field.field), schematic.nested_attributes)
        else:
            if schematic.is_list:
                return list(value)
            else:
                return list(value)[0]



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path
from fnmatch import fnmatch
from typing import IO, Any, Callable

import yaml
from fameio.source.logs import log_and_raise_critical, log
from fameio.source.path_resolver import PathResolver

DISABLING_YAML_FILE_PREFIX = "IGNORE_"

CRIT_ERR_NO_YAML_GIVEN = "Please provide a valid YAML file. Received `-f/--file {}` instead."


class Args:
    def __init__(self, file_string, node_string):
        self.file_string = file_string
        self.node_string = node_string


def read_args(loader, args):
    """Returns two Strings to be interpreted as files to be read and a sub-node_address from the given `args`"""
    node_string = ""
    file_string = None
    if isinstance(args, yaml.nodes.ScalarNode):
        file_string = loader.construct_scalar(args)
        log().debug("Found instance `ScalarNode` in {}".format(file_string))
    elif isinstance(args, yaml.nodes.SequenceNode):
        argument_list = loader.construct_sequence(args)
        if len(argument_list) not in [1, 2]:
            log_and_raise_critical("!include supports but one or two arguments in list")
        elif len(argument_list) == 2:
            node_string = argument_list[1]
        file_string = argument_list[0]
        log().debug("Found instance `SequenceNode` in {}".format(file_string))
    elif isinstance(args, yaml.nodes.MappingNode):
        argument_map = loader.construct_mapping(args)
        for key, value in argument_map.items():
            if key.lower() == "file":
                file_string = value
            elif key.lower() == "node":
                node_string = value
            else:
                log_and_raise_critical("!include supports only keys 'file' and 'node'")
        if not file_string:
            log_and_raise_critical("!include: file must be specified.")
    else:
        log_and_raise_critical("YAML node type not implemented: {}".format(args))
    return Args(file_string, node_string)


def split_nodes(node_string):
    """Returns a list of nodes created from the given `node_string`"""
    log().debug("Splitting given node_string `{}`".format(node_string))
    return node_string.split(":")


class FameYamlLoader(yaml.SafeLoader):
    """Custom YAML Loader for `!include` constructor"""

    def __init__(self, stream: IO, path_resolver=PathResolver()) -> None:
        log().debug("Initialize custom YAML loader")
        self._path_resolver = path_resolver
        try:
            self._root_path = os.path.split(stream.name)[0]
        except AttributeError:
            self._root_path = os.path.curdir
        super().__init__(stream)

    @property
    def root_path(self) -> str:
        return self._root_path

    @property
    def path_resolver(self) -> PathResolver:
        return self._path_resolver


def make_yaml_loader_builder(
    path_resolver: PathResolver,
) -> Callable[[Any], FameYamlLoader]:
    """Utility function used to control the creation of a FameYamlLoader by the YAML library."""
    return lambda stream: FameYamlLoader(stream, path_resolver)


def resolve_imported_path(loader: FameYamlLoader, included_path: str):
    """
    Returns a list of file paths matching the given `included_path` based on path resolution performed by the loader.

    Ignores the files starting with `DISABLING_YAML_FILE_PREFIX`
    """
    file_list = loader.path_resolver.resolve_yaml_imported_file_pattern(loader.root_path, included_path)

    ignore_filter = "*" + DISABLING_YAML_FILE_PREFIX + "*"
    cleaned_file_list = []
    for file in file_list:
        if fnmatch(file, ignore_filter):
            log().debug("Ignoring file {} due to prefix {}".format(file, DISABLING_YAML_FILE_PREFIX))
        else:
            cleaned_file_list.append(file)
    if not cleaned_file_list:
        log_and_raise_critical("Failed to find any file matching the `!include` directive `{}`".format(included_path))
    log().debug("Collected file(s) `{}` from given included path `{}`".format(cleaned_file_list, included_path))
    return cleaned_file_list


def read_data_from_file(file, node_address, path_resolver: PathResolver):
    """Returns data of the specified `node_address` from the specified `file`"""
    data = yaml.load(file, make_yaml_loader_builder(path_resolver))
    for node in node_address:
        if node:
            try:
                data = data[node]
            except KeyError:
                log_and_raise_critical("'!include_node [{}, {}]': Cannot find '{}'.".format(file, node_address, node))
    log().debug("Searched file `{}` for node `{}`".format(file, node_address))
    return data


def join_data(new_data, previous_data):
    """Joins data from multiple files if both are in list format, otherwise returns"""
    if not previous_data:
        return new_data
    if isinstance(new_data, list) and isinstance(previous_data, list):
        previous_data.extend(new_data)
        return previous_data
    else:
        log_and_raise_critical("!include can only combine list-like elements from multiple files!")


def construct_include(loader: FameYamlLoader, args: yaml.Node) -> Any:
    """
    Loads one or many YAML file(s) with specifications provided in `args` in different formats
    To load all content of a specified file, use:
        !include "path/to/file.yaml"
    To load only specific content (e.g. data of "Super:Sub:Node") from a given file, use:
        !include ["path/to/file.yml", "Super:Sub:Node"]
    For a slightly more verbose version of the above commands, use a dictionary argument:
        !include {"file":"path/to/file.yml", "node": "Super:Sub:Node"}

    Instead of "path/to/file.yaml" one can also use asterisks to select multiple files, e.g. "path/to/files/*.yaml"
    The given file path is either
      * relative to the path of the including YAML file, if it starts with a character other than "/"
      * an absolute path if its starts with "/"
    """
    args = read_args(loader, args)
    nodes = split_nodes(args.node_string)
    files = resolve_imported_path(loader, args.file_string)

    joined_data = None
    for file_name in files:
        with open(file_name, "r") as open_file:
            data = read_data_from_file(open_file, nodes, loader.path_resolver)
            joined_data = join_data(data, joined_data)
    log().debug("Joined all files `{}` to joined data `{}`".format(files, joined_data))
    return joined_data


FameYamlLoader.add_constructor("!include", construct_include)


def load_yaml(yaml_file_path: Path, path_resolver=PathResolver()):
    """Loads the yaml file from given `yaml_file_path` and returns its content"""
    log().info("Loading yaml from {}".format(yaml_file_path))
    with open(yaml_file_path, "r") as configfile:
        data = yaml.load(configfile, make_yaml_loader_builder(path_resolver))
    return data


def check_for_yaml_file_type(yaml_file_path: Path) -> None:
    """Raises Exception if given `yaml_file_path` is not a valid YAML file"""
    suffix = yaml_file_path.suffix.lower()
    if suffix != ".yaml" and suffix != ".yml":
        log_and_raise_critical(CRIT_ERR_NO_YAML_GIVEN.format(yaml_file_path))



#!/usr/bin/env python

import sys
from pathlib import Path

from fameio.source.cli.convert_results import handle_args, CLI_DEFAULTS as DEFAULT_CONFIG
from fameio.source.cli.options import Options
from fameio.source.cli.parser import update_default_config
from fameio.source.logs import log_and_raise_critical, fameio_logger, log
from fameio.source.results.agent_type import AgentTypeLog
from fameio.source.results.conversion import apply_time_option, apply_time_merging
from fameio.source.results.csv_writer import CsvWriter
from fameio.source.results.data_transformer import DataTransformer
from fameio.source.results.input_dao import InputDao
from fameio.source.results.output_dao import OutputDAO
from fameio.source.results.reader import Reader
from fameio.source.results.yaml_writer import data_to_yaml_file

ERR_MEMORY_ERROR = "Out of memory. Try using `-m` or `--memory-saving` option."
ERR_MEMORY_SEVERE = "Out of memory despite memory-saving mode. Reduce output interval in `FAME-Core` and rerun model"


def run(config: dict = None) -> None:
    """Reads file in protobuf format for configures FILE and extracts its content to .csv file(s)"""
    config = update_default_config(config, DEFAULT_CONFIG)
    fameio_logger(log_level_name=config[Options.LOG_LEVEL], file_name=config[Options.LOG_FILE])

    file_path = config[Options.FILE]
    output_writer = CsvWriter(config[Options.OUTPUT], Path(file_path), config[Options.SINGLE_AGENT_EXPORT])
    file_stream = open(Path(file_path), "rb")

    if config[Options.MEMORY_SAVING]:
        log().info("Memory saving mode enabled: Disable on conversion of small files for performance improvements.")

    log().info("Reading and extracting data...")
    reader = Reader.get_reader(file=file_stream, read_single=config[Options.MEMORY_SAVING])
    agent_type_log = AgentTypeLog(requested_agents=config[Options.AGENT_LIST])
    data_transformer = DataTransformer.build(config[Options.RESOLVE_COMPLEX_FIELD])
    try:
        input_dao = InputDao()
        while data_storages := reader.read():
            if config[Options.INPUT_RECOVERY]:
                input_dao.store_inputs(data_storages)
            output = OutputDAO(data_storages, agent_type_log)
            for agent_name in output.get_sorted_agents_to_extract():
                log().debug(f"Extracting data for {agent_name}...")
                data_frames = output.get_agent_data(agent_name, data_transformer)
                apply_time_merging(data_frames, config[Options.TIME_MERGING])
                apply_time_option(data_frames, config[Options.TIME])
                log().debug(f"Writing data for {agent_name}...")
                output_writer.write_to_files(agent_name, data_frames)

        if config[Options.INPUT_RECOVERY]:
            log().info("Recovering inputs...")
            timeseries, scenario = input_dao.recover_inputs()
            series_writer = CsvWriter(Path(config[Options.OUTPUT], "./recovered"), Path("./"), False)
            series_writer.write_time_series_to_disk(timeseries)
            data_to_yaml_file(scenario.to_dict(), Path(config[Options.OUTPUT], "./recovered/scenario.yaml"))

        log().info("Data conversion completed.")
    except MemoryError:
        log_and_raise_critical(ERR_MEMORY_SEVERE if Options.MEMORY_SAVING else ERR_MEMORY_ERROR)

    file_stream.close()
    if not agent_type_log.has_any_agent_type():
        log().error("Provided file did not contain any output data.")


if __name__ == "__main__":
    run_config = handle_args(sys.argv[1:])
    run(run_config)



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from .agent import Agent
from .attribute import Attribute
from .contract import Contract
from .exception import ScenarioException
from .fameiofactory import FameIOFactory
from .generalproperties import GeneralProperties
from .scenario import Scenario



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from .agenttype import AgentType
from .attribute import AttributeSpecs, AttributeType
from .exception import SchemaException
from .schema import Schema



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0


class SchemaException(Exception):
    """An exception that occurred while parsing a Schema"""

    pass



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC
from builtins import staticmethod
from typing import List, Dict, Tuple, Union, Optional

import pandas as pd
from fameprotobuf.Services_pb2 import Output
from pandas import DataFrame

from fameio.source.cli.options import ResolveOptions
from fameio.source.results.agent_type import AgentType

INDEX = ("AgentId", "TimeStep")


class DataTransformer(ABC):
    """Extracts and provides series data from parsed and processed output files for requested agents"""

    MODES = {
        ResolveOptions.IGNORE: lambda: DataTransformerIgnore(),
        ResolveOptions.SPLIT: lambda: DataTransformerSplit(),
    }
    SIMPLE_COLUMN_INDEX = -1

    @staticmethod
    def build(complex_column_mode: ResolveOptions) -> DataTransformer:
        return DataTransformer.MODES[complex_column_mode]()

    def extract_agent_data(
        self, series: List[Output.Series], agent_type: AgentType
    ) -> Dict[Optional[str], pd.DataFrame]:
        """
        Returns dict of DataFrame(s) containing all data from given `series` of given `agent_type`.
        When ResolveOption is `SPLIT`, the dict maps each complex column's name to the associated DataFrame.
        In any case, the dict maps `None` to a DataFrame with the content of all simple column / merged columns.
        """
        container = self._extract_agent_data(series, agent_type)
        data_frames = {}
        for column_id, agent_data in container.items():
            data_frame = DataFrame.from_dict(agent_data, orient="index")
            column_name = agent_type.get_column_name_for_id(column_id)
            if column_id == DataTransformer.SIMPLE_COLUMN_INDEX:
                data_frame.rename(columns=self._get_column_map(agent_type), inplace=True)
                index = INDEX
                data_frame = data_frame.loc[:, agent_type.get_simple_column_mask()]
            else:
                data_frame.rename(columns={0: column_name}, inplace=True)
                index = INDEX + agent_type.get_inner_columns(column_id)

            if not data_frame.empty:
                data_frame.index = pd.MultiIndex.from_tuples(data_frame.index)
                data_frame.rename_axis(index, inplace=True)
            data_frames[column_name] = data_frame
        return data_frames

    def _extract_agent_data(
        self, series: List[Output.Series], agent_type: AgentType
    ) -> Dict[int, Dict[Tuple, List[Union[float, None, str]]]]:
        """Returns mapping of (agentId, timeStep) to fixed-length list of all output columns for given `class_name`"""
        container = DataTransformer._create_container(agent_type)
        mask_simple = agent_type.get_simple_column_mask()
        while series:
            self._add_series_data(series.pop(), mask_simple, container)
        filled_columns = {index: column_data for index, column_data in container.items() if len(column_data) > 0}
        return filled_columns

    @staticmethod
    def _create_container(agent_type: AgentType) -> Dict[int, Dict]:
        """Returns map of complex columns IDs to an empty dict, and one more for the remaining simple columns"""
        field_ids = agent_type.get_complex_column_ids().union([DataTransformer.SIMPLE_COLUMN_INDEX])
        return {field_id: {} for field_id in field_ids}

    def _add_series_data(
        self,
        series: Output.Series,
        mask_simple: List[bool],
        container: Dict[int, Dict[Tuple, List[Union[float, None, str]]]],
    ) -> None:
        """Adds data from given `series` to specified `container` dict as list"""
        empty_list = [None] * len(mask_simple)
        for line in series.line:
            index = (series.agentId, line.timeStep)
            simple_values = empty_list.copy()
            for column in line.column:
                if mask_simple[column.fieldId]:
                    simple_values[column.fieldId] = column.value
                else:
                    self._merge_complex_column(column, simple_values)
                    self._store_complex_values(column, container, index)
            container[DataTransformer.SIMPLE_COLUMN_INDEX][index] = simple_values

    @staticmethod
    def _merge_complex_column(column: Output.Series.Line.Column, values: List) -> None:
        """Does not merge complex column data"""
        pass

    @staticmethod
    def _store_complex_values(column: Output.Series.Line.Column, container: Dict[int, Dict], index: Tuple) -> None:
        """Does not store complex column data"""
        pass

    @staticmethod
    def _get_column_map(agent_type: AgentType) -> Dict[int, str]:
        """Returns mapping of simple column IDs to their name for given `agent_type`"""
        return agent_type.get_simple_column_map()


class DataTransformerIgnore(DataTransformer):
    """Ignores complex columns on output"""


class DataTransformerSplit(DataTransformer):
    @staticmethod
    def _store_complex_values(column: Output.Series.Line.Column, container: Dict[int, Dict], base_index: Tuple) -> None:
        """Adds inner data from `column` to given `container` - split by column type"""
        for entry in column.entry:
            index = base_index + tuple(entry.indexValue)
            container[column.fieldId][index] = entry.value



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import argparse
from enum import Enum, auto


class ParsableEnum(Enum):
    """Extend this to create an enum that can be parsed with argparse"""

    @classmethod
    def instantiate(cls, name: str) -> Enum:
        try:
            return cls[name]
        except KeyError:
            raise argparse.ArgumentTypeError(f"'{name}' is not a valid option")

    def __str__(self):
        return self.name


class Options(Enum):
    """Specifies command line configuration options"""

    FILE = auto()
    LOG_LEVEL = auto()
    LOG_FILE = auto()
    OUTPUT = auto()
    AGENT_LIST = auto()
    SINGLE_AGENT_EXPORT = auto()
    MEMORY_SAVING = auto()
    RESOLVE_COMPLEX_FIELD = auto()
    TIME = auto()
    TIME_MERGING = auto()
    INPUT_RECOVERY = auto()


class TimeOptions(ParsableEnum, Enum):
    """Specifies options for conversion of time in output"""

    INT = auto()
    UTC = auto()
    FAME = auto()


class ResolveOptions(ParsableEnum, Enum):
    """Specifies options for resolving complex fields in output files"""

    IGNORE = auto()
    SPLIT = auto()


class MergingOptions(Enum):
    """Specifies options for merging TimeSteps"""

    FOCAL_POINT = auto()
    STEPS_BEFORE = auto()
    STEPS_AFTER = auto()



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import ast
from typing import Dict, Optional

from fameio.source.logs import log_error_and_raise, log
from fameio.source.schema.agenttype import AgentType
from fameio.source.schema.exception import SchemaException
from fameio.source.schema.java_packages import JavaPackages
from fameio.source.tools import keys_to_lower


class Schema:
    """Definition of a schema"""

    _ERR_AGENT_TYPES_MISSING = "Required keyword `AgentTypes` missing in Schema."
    _ERR_AGENT_TYPES_EMPTY = "`AgentTypes` must not be empty - at least one type of agent is required."
    _WARN_MISSING_PACKAGES = "No `JavaPackages` defined Schema. Future versions of fameio will require it."

    _KEY_AGENT_TYPE = "AgentTypes".lower()
    _KEY_PACKAGES = "JavaPackages".lower()

    def __init__(self, definitions: dict):
        self._original_input_dict = definitions
        self._agent_types = {}
        self._packages: Optional[JavaPackages] = None

    @classmethod
    def from_dict(cls, definitions: dict) -> Schema:
        """Load given dictionary `definitions` into a new Schema"""
        definitions = keys_to_lower(definitions)
        if Schema._KEY_AGENT_TYPE not in definitions:
            log_error_and_raise(SchemaException(Schema._ERR_AGENT_TYPES_MISSING))
        schema = cls(definitions)
        agent_types = definitions[Schema._KEY_AGENT_TYPE]
        if len(agent_types) == 0:
            log_error_and_raise(SchemaException(Schema._ERR_AGENT_TYPES_EMPTY))

        for agent_type_name, agent_definition in agent_types.items():
            agent_type = AgentType.from_dict(agent_type_name, agent_definition)
            schema._agent_types[agent_type_name] = agent_type

        if Schema._KEY_PACKAGES in definitions:
            schema._packages = JavaPackages.from_dict(definitions[Schema._KEY_PACKAGES])
        else:
            log().warning(Schema._WARN_MISSING_PACKAGES)
            schema._packages = JavaPackages()
        return schema

    @classmethod
    def from_string(cls, definitions: str) -> Schema:
        """Load given string `definitions` into a new Schema"""
        return cls.from_dict(ast.literal_eval(definitions))

    def to_dict(self) -> dict:
        """Serializes the schema content to a dict"""
        return self._original_input_dict

    def to_string(self) -> str:
        """Returns a string representation of the Schema of which the class can be rebuilt"""
        return repr(self.to_dict())

    @property
    def agent_types(self) -> Dict[str, AgentType]:
        """Returns all the agent types by their name"""
        return self._agent_types

    @property
    def packages(self) -> JavaPackages:
        """Returns JavaPackages, i.e. names where model classes are defined in"""
        return self._packages



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from typing import List, Dict, Set, Union, Tuple

from fameprotobuf.Services_pb2 import Output


class AgentType:
    """Provides information derived from an underlying protobuf AgentType"""

    def __init__(self, agent_type: Output.AgentType) -> None:
        self._agent_type = agent_type

    def get_simple_column_map(self) -> Dict[int, str]:
        """Returns dictionary of simple column IDs mapped to their name - ignoring complex columns"""
        return {field.fieldId: field.fieldName for field in self._agent_type.field if len(field.indexName) == 0}

    def get_merged_column_map(self) -> Dict[int, str]:
        """Returns dictionary of all column IDs mapped to their name merged with names of inner complex columns"""
        column_names = {}
        for field in self._agent_type.field:
            if len(field.indexName) == 0:
                column_names[field.fieldId] = field.fieldName
            else:
                column_names[field.fieldId] = f"{field.fieldName}_({tuple(field.indexName)}, value)"
        return column_names

    def get_simple_column_mask(self) -> List[bool]:
        """Returns list of bool - where an entry is True if the output column with the same index is not complex"""
        return [True if len(field.indexName) == 0 else False for field in self._agent_type.field]

    def get_complex_column_ids(self) -> Set[int]:
        """Returns set of IDs for complex columns, ignoring simple columns"""
        return set([field.fieldId for field in self._agent_type.field if len(field.indexName) > 0])

    def get_column_name_for_id(self, column_index: int) -> Union[str, None]:
        """Returns name of column by given `column_index` or None, if column is not present"""
        if 0 <= column_index < len(self._agent_type.field):
            return self._agent_type.field[column_index].fieldName
        else:
            return None

    def get_inner_columns(self, column_index: int) -> Tuple[str]:
        """Returns tuple of inner column names for complex column with given `column_index`"""
        return tuple(self._agent_type.field[column_index].indexName)

    def get_class_name(self) -> str:
        """Returns name of class of wrapped agent type"""
        return self._agent_type.className


class AgentTypeLog:
    """Stores data about collected agent types"""

    _ERR_AGENT_TYPE_MISSING = "Requested AgentType `{}` not found."
    _ERR_DOUBLE_DEFINITION = "Just one definition allowed per AgentType. Found multiple for {}. File might be corrupt."

    def __init__(self, requested_agents: List[str]) -> None:
        self._requested_agents = [agent.upper() for agent in requested_agents] if requested_agents else None
        self._requested_agent_types = {}

    def update_agents(self, new_types: Dict[str, Output.AgentType]) -> None:
        """Saves new `agent_types` (if any) contained in given `output` if requested for extraction"""
        if new_types:
            if self._requested_agents:
                new_types = {
                    agent_name: agent_type
                    for agent_name, agent_type in new_types.items()
                    if agent_name.upper() in self._requested_agents
                }
            for agent_name in self._requested_agent_types.keys():
                if agent_name in new_types:
                    raise Exception(self._ERR_DOUBLE_DEFINITION.format(agent_name))
            self._requested_agent_types.update(new_types)

    def has_any_agent_type(self) -> bool:
        """Returns True if any agent type was registered so far present"""
        return len(self._requested_agent_types) > 0

    def get_agent_type(self, agent_name: str) -> AgentType:
        """Returns `AgentType` of given agent `name`"""
        if agent_name not in self._requested_agent_types:
            raise Exception(self._ERR_AGENT_TYPE_MISSING.format(agent_name))
        return AgentType(self._requested_agent_types[agent_name])

    def is_requested(self, agent_name: str) -> bool:
        """Returns True if given agent_name is known and requested"""
        return agent_name in self._requested_agent_types



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import math
from collections import Counter
from typing import Any, Dict, List

from fameio.source import PathResolver
from fameio.source.logs import log_error_and_raise, log
from fameio.source.scenario import Agent, Attribute, Contract, Scenario
from fameio.source.schema.agenttype import AgentType
from fameio.source.schema.attribute import AttributeSpecs, AttributeType
from fameio.source.schema.schema import Schema
from fameio.source.series import TimeSeriesManager, TimeSeriesException
from fameio.source.time import FameTime


class ValidationException(Exception):
    """Indicates an error occurred during validation of any data with a connected schema"""

    pass


class SchemaValidator:
    """Handles validation of scenarios based on a connected `schema`"""

    _AGENT_ID_NOT_UNIQUE = "Agent ID(s) not unique: '{}'."
    _AGENT_TYPE_UNKNOWN = "Agent type '{}' not declared in Schema."
    _ATTRIBUTE_UNKNOWN = "Attribute '{}' not declared in Schema."
    _TYPE_NOT_IMPLEMENTED = "Validation not implemented for AttributeType '{}'."
    _INCOMPATIBLE = "Value '{}' incompatible with {} of Attribute '{}'."
    _DISALLOWED = "Value '{}' not in list of allowed values of Attribute '{}'"
    _AGENT_MISSING = "Contract mentions Agent with ID '{}' but Agent was not declared."
    _PRODUCT_MISSING = "Product '{}' not declared in Schema for AgentType '{}'."
    _KEY_MISSING = "Required key '{}' missing in dictionary '{}'."
    _ATTRIBUTE_MISSING = "Mandatory attribute '{}' is missing."
    _DEFAULT_IGNORED = "Optional Attribute: '{}': not specified - provided Default ignored for optional Attributes."
    _OPTIONAL_MISSING = "Optional Attribute: '{}': not specified."
    _IS_NO_LIST = "Attribute '{}' is list but assigned value '{}' is not a list."
    _TIME_SERIES_INVALID = "Timeseries at '{}' is invalid."
    _MISSING_CONTRACTS_FOR_AGENTS = "No contracts defined for Agent '{}' of type '{}'"

    @staticmethod
    def validate_scenario_and_timeseries(
        scenario: Scenario, path_resolver: PathResolver = PathResolver()
    ) -> TimeSeriesManager:
        """
        Validates the given `scenario` and its timeseries using given `path_resolver`
        Raises an exception if schema requirements are not met or timeseries data are erroneous.

        Args:
            scenario: to be validated against the encompassed schema
            path_resolver: to resolve paths of timeseries

        Returns:
            a new TimeSeriesManager initialised with validated time series from scenario
        Raises:
            ValidationException: if an error in the scenario or in timeseries is spotted
        """
        schema = scenario.schema
        agents = scenario.agents
        timeseries_manager = TimeSeriesManager(path_resolver)

        SchemaValidator.ensure_unique_agent_ids(agents)
        for agent in agents:
            SchemaValidator.ensure_agent_and_timeseries_are_valid(agent, schema, timeseries_manager)

        agent_types_by_id = {agent.id: agent.type_name for agent in agents}
        for contract in scenario.contracts:
            SchemaValidator.ensure_is_valid_contract(contract, schema, agent_types_by_id)

        return timeseries_manager

    @staticmethod
    def ensure_unique_agent_ids(agents: List[Agent]) -> None:
        """Raises exception if any id for given `agents` is not unique"""
        list_of_ids = [agent.id for agent in agents]
        non_unique_ids = [agent_id for agent_id, count in Counter(list_of_ids).items() if count > 1]
        if non_unique_ids:
            log_error_and_raise(ValidationException(SchemaValidator._AGENT_ID_NOT_UNIQUE.format(non_unique_ids)))

    @staticmethod
    def ensure_agent_and_timeseries_are_valid(agent: Agent, schema: Schema, timeseries_manager: TimeSeriesManager):
        """Validates given `agent` against `schema` plus loads and validates its timeseries"""
        SchemaValidator.ensure_agent_type_in_schema(agent, schema)
        SchemaValidator.ensure_is_valid_agent(agent, schema)
        SchemaValidator.load_and_validate_timeseries(agent, schema, timeseries_manager)

    @staticmethod
    def ensure_agent_type_in_schema(agent: Agent, schema: Schema) -> None:
        """Raises exception if type for given `agent` is not specified in given `schema`"""
        if agent.type_name not in schema.agent_types:
            log_error_and_raise(ValidationException(SchemaValidator._AGENT_TYPE_UNKNOWN.format(agent.type_name)))

    @staticmethod
    def ensure_is_valid_agent(agent: Agent, schema: Schema) -> None:
        """Raises an exception if given `agent` does not meet the specified `schema` requirements"""
        scenario_attributes = agent.attributes
        schema_attributes = SchemaValidator._get_agent(schema, agent.type_name).attributes
        SchemaValidator._ensure_mandatory_present(scenario_attributes, schema_attributes)
        SchemaValidator._ensure_attributes_exist(scenario_attributes, schema_attributes)
        SchemaValidator._ensure_value_and_type_match(scenario_attributes, schema_attributes)

    @staticmethod
    def _get_agent(schema: Schema, name: str) -> AgentType:
        """Returns agent specified by `name` or raises Exception if this agent is not present in given `schema`"""
        if name in schema.agent_types:
            return schema.agent_types[name]
        else:
            log_error_and_raise(ValidationException(SchemaValidator._AGENT_TYPE_UNKNOWN.format(name)))

    @staticmethod
    def _ensure_mandatory_present(attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs]) -> None:
        """
        Raises Exception if in given list of `specifications` at least one item is mandatory,
        provides no defaults and is not contained in given `attributes` dictionary
        """
        for name, specification in specifications.items():
            if name not in attributes:
                if specification.is_mandatory:
                    if not specification.has_default_value:
                        log_error_and_raise(
                            ValidationException(SchemaValidator._ATTRIBUTE_MISSING.format(specification.full_name))
                        )
                else:
                    if specification.has_default_value:
                        log().warning(SchemaValidator._DEFAULT_IGNORED.format(specification.full_name))
                    else:
                        log().info(SchemaValidator._OPTIONAL_MISSING.format(specification.full_name))
            if name in attributes and specification.has_nested_attributes:
                attribute = attributes[name]
                if specification.is_list:
                    for entry in attribute.nested_list:
                        SchemaValidator._ensure_mandatory_present(entry, specification.nested_attributes)
                else:
                    SchemaValidator._ensure_mandatory_present(attribute.nested, specification.nested_attributes)

    @staticmethod
    def _ensure_attributes_exist(attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs]) -> None:
        """Raises exception any entry of given `attributes` has no corresponding type `specification`"""
        for name, attribute in attributes.items():
            if name not in specifications:
                log_error_and_raise(ValidationException(SchemaValidator._ATTRIBUTE_UNKNOWN.format(attribute)))
            if attribute.has_nested:
                specification = specifications[name]
                SchemaValidator._ensure_attributes_exist(attribute.nested, specification.nested_attributes)
            if attribute.has_nested_list:
                specification = specifications[name]
                for entry in attribute.nested_list:
                    SchemaValidator._ensure_attributes_exist(entry, specification.nested_attributes)

    @staticmethod
    def _ensure_value_and_type_match(
        attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs]
    ) -> None:
        """Raises exception if in given list of `attributes` its value does not match associated type `specification`"""
        for name, attribute in attributes.items():
            specification = specifications[name]
            if attribute.has_value:
                value = attribute.value
                type_spec = specification.attr_type
                if not SchemaValidator._is_compatible(specification, value):
                    message = SchemaValidator._INCOMPATIBLE.format(value, type_spec, specification.full_name)
                    log_error_and_raise(ValidationException(message))
                if not SchemaValidator._is_allowed_value(specification, value):
                    message = SchemaValidator._DISALLOWED.format(value, specification.full_name)
                    log_error_and_raise(ValidationException(message))
            if attribute.has_nested:
                SchemaValidator._ensure_value_and_type_match(attribute.nested, specification.nested_attributes)
            if attribute.has_nested_list:
                for entry in attribute.nested_list:
                    SchemaValidator._ensure_value_and_type_match(entry, specification.nested_attributes)

    @staticmethod
    def _is_compatible(specification: AttributeSpecs, value_or_values: Any) -> bool:
        """Returns True if given `value_or_values` is compatible to specified `attribute_type` and `should_be_list`"""
        is_list = isinstance(value_or_values, list)
        attribute_type = specification.attr_type
        if specification.is_list:
            if not is_list:
                log().warning(SchemaValidator._IS_NO_LIST.format(specification.full_name, value_or_values))
                return SchemaValidator._is_compatible_value(attribute_type, value_or_values)
            for value in value_or_values:
                if not SchemaValidator._is_compatible_value(attribute_type, value):
                    return False
            return True
        else:
            return (not is_list) and SchemaValidator._is_compatible_value(attribute_type, value_or_values)

    @staticmethod
    def _is_compatible_value(attribute_type: AttributeType, value) -> bool:
        """Returns True if given single value is compatible to specified `attribute_type` and is not a NaN float"""
        if attribute_type is AttributeType.INTEGER:
            if isinstance(value, int):
                return -2147483648 < value < 2147483647
            return False
        if attribute_type is AttributeType.LONG:
            return isinstance(value, int)
        elif attribute_type is AttributeType.DOUBLE:
            return isinstance(value, (int, float)) and not math.isnan(value)
        elif attribute_type in (AttributeType.ENUM, AttributeType.STRING):
            return isinstance(value, str)
        elif attribute_type is AttributeType.TIME_STAMP:
            return FameTime.is_fame_time_compatible(value)
        elif attribute_type is AttributeType.TIME_SERIES:
            return isinstance(value, (str, int)) or (isinstance(value, float) and not math.isnan(value))
        else:
            log_error_and_raise(ValidationException(SchemaValidator._TYPE_NOT_IMPLEMENTED.format(attribute_type)))

    @staticmethod
    def _is_allowed_value(attribute: AttributeSpecs, value) -> bool:
        """Returns True if `value` matches an entry of given `Attribute`'s value list or if this list is empty"""
        if not attribute.values:
            return True
        else:
            return value in attribute.values

    @staticmethod
    def load_and_validate_timeseries(agent: Agent, schema: Schema, timeseries_manager: TimeSeriesManager) -> None:
        """
        Loads all timeseries specified in given `schema` of given `agent` into given `timeseries_manager`

        Args:
            agent: definition in scenario
            schema: schema encompassed in scenario
            timeseries_manager: to be filled with timeseries

        Raises:
            ValidationException: if timeseries is not found, ill-formatted or invalid
        """
        scenario_attributes = agent.attributes
        schema_attributes = SchemaValidator._get_agent(schema, agent.type_name).attributes
        SchemaValidator._ensure_valid_timeseries(scenario_attributes, schema_attributes, timeseries_manager)

    @staticmethod
    def _ensure_valid_timeseries(
        attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs], manager: TimeSeriesManager
    ) -> None:
        """Recursively searches for time_series in agent attributes and registers them at given `manager`"""
        for name, attribute in attributes.items():
            specification = specifications[name]
            if attribute.has_value:
                attribute_type = specification.attr_type
                if attribute_type is AttributeType.TIME_SERIES:
                    try:
                        manager.register_and_validate(attribute.value)
                    except TimeSeriesException as e:
                        message = SchemaValidator._TIME_SERIES_INVALID.format(specification.full_name)
                        log_error_and_raise(ValidationException(message, e))
            if attribute.has_nested:
                SchemaValidator._ensure_valid_timeseries(attribute.nested, specification.nested_attributes, manager)
            if attribute.has_nested_list:
                for entry in attribute.nested_list:
                    SchemaValidator._ensure_valid_timeseries(entry, specification.nested_attributes, manager)

    @staticmethod
    def ensure_is_valid_contract(contract: Contract, schema: Schema, agent_types_by_id: Dict[int, str]) -> None:
        """Raises exception if given `contract` does not meet the `schema`'s requirements, using `agent_types_by_id`"""
        sender_id = contract.sender_id
        if sender_id not in agent_types_by_id:
            log_error_and_raise(ValidationException(SchemaValidator._AGENT_MISSING.format(sender_id)))
        if contract.receiver_id not in agent_types_by_id:
            log_error_and_raise(ValidationException(SchemaValidator._AGENT_MISSING.format(contract.receiver_id)))
        sender_type_name = agent_types_by_id[sender_id]
        if sender_type_name not in schema.agent_types:
            log_error_and_raise(ValidationException(SchemaValidator._AGENT_TYPE_UNKNOWN.format(sender_type_name)))
        sender_type = schema.agent_types[sender_type_name]
        product = contract.product_name
        if product not in sender_type.products:
            log_error_and_raise(ValidationException(SchemaValidator._PRODUCT_MISSING.format(product, sender_type_name)))

    @staticmethod
    def check_agents_have_contracts(scenario: Scenario) -> None:
        """Raises warning for each agent without any assigned contract"""
        senders = [contract.sender_id for contract in scenario.contracts]
        receivers = [contract.receiver_id for contract in scenario.contracts]
        active_agents = set(senders + receivers)
        inactive_agents = {agent.id: agent.type_name for agent in scenario.agents if agent.id not in active_agents}

        if inactive_agents:
            for agent_id, agent_name in inactive_agents.items():
                log().warning(SchemaValidator._MISSING_CONTRACTS_FOR_AGENTS.format(agent_id, agent_name))



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import List, Dict

from fameio.source.logs import log_error_and_raise
from fameio.source.schema import SchemaException
from fameio.source.tools import keys_to_lower


class JavaPackages:
    """Schema definitions for Java package names in which model classes reside"""

    _ERR_MISSING_AGENTS = "JavaPackages requires non-empty list for `Agents`. Key was missing or list was empty."
    _ERR_MISSING_DATA_ITEMS = "JavaPackages requires non-empty list for `DataItems`. Key was missing or list was empty."
    _ERR_MISSING_PORTABLES = "JavaPackages require non-empty list for `Portables`. Key was missing or list was empty."

    _KEY_AGENT = "Agents".lower()
    _KEY_DATA_ITEM = "DataItems".lower()
    _KEY_PORTABLE = "Portables".lower()

    def __init__(self):
        self._agents: List[str] = []
        self._data_items: List[str] = []
        self._portables: List[str] = []

    @classmethod
    def from_dict(cls, definitions: Dict[str, List[str]]) -> JavaPackages:
        """
        Creates JavaPackages from a dictionary representation

        Args:
            definitions: dictionary representation of JavaPackages

        Returns:
            new instance of JavaPackages
        """
        java_packages = cls()
        definitions = keys_to_lower(definitions)

        java_packages._agents = definitions.get(JavaPackages._KEY_AGENT, [])
        java_packages._data_items = definitions.get(JavaPackages._KEY_DATA_ITEM, [])
        java_packages._portables = definitions.get(JavaPackages._KEY_PORTABLE, [])

        if not java_packages._agents:
            log_error_and_raise(SchemaException(JavaPackages._ERR_MISSING_AGENTS))
        if not java_packages._data_items:
            log_error_and_raise(SchemaException(JavaPackages._ERR_MISSING_DATA_ITEMS))
        if not java_packages._portables:
            log_error_and_raise(SchemaException(JavaPackages._ERR_MISSING_PORTABLES))

        return java_packages

    @property
    def agents(self) -> List[str]:
        """Return list of java package names that contain the model's Agents"""
        return self._agents

    @property
    def data_items(self) -> List[str]:
        """Return list of java package names that contain the model's DataItems"""
        return self._data_items

    @property
    def portables(self) -> List[str]:
        """Return list of java package names that contain the model's Portables"""
        return self._portables



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Dict, Union

import pandas as pd

from fameio.source.logs import log
from fameio.source.results.data_transformer import INDEX
from fameio.source.series import TimeSeriesManager
from fameio.source.tools import ensure_path_exists


class CsvWriter:
    """Writes dataframes to different csv files"""

    _INFO_USING_PATH = "Using specified output path: {}"
    _INFO_USING_DERIVED_PATH = "No output path specified - writing to new local folder: {}"

    def __init__(self, config_output: Path, input_file_path: Path, single_export: bool) -> None:
        self._single_export = single_export
        self._output_folder = self._get_output_folder_name(config_output, input_file_path)
        self._files = {}
        self._create_output_folder()

    @staticmethod
    def _get_output_folder_name(config_output: Path, input_file_path: Path) -> Path:
        """Returns name of the output folder derived either from the specified `config_output` or `input_file_path`"""
        if config_output:
            output_folder_name = config_output
            log().info(CsvWriter._INFO_USING_PATH.format(config_output))
        else:
            output_folder_name = input_file_path.stem
            log().info(CsvWriter._INFO_USING_DERIVED_PATH.format(output_folder_name))
        return Path(output_folder_name)

    def _create_output_folder(self) -> None:
        """Creates output folder if not yet present"""
        log().debug("Creating output folder if required...")
        if not self._output_folder.is_dir():
            self._output_folder.mkdir(parents=True)

    def write_to_files(self, agent_name: str, data: Dict[Union[None, str], pd.DataFrame]) -> None:
        """Writes `data` for given `agent_name` to .csv file(s)"""
        for column_name, column_data in data.items():
            column_data.sort_index(inplace=True)
            if self._single_export:
                for agent_id, agent_data in column_data.groupby(INDEX[0]):
                    identifier = self._get_identifier(agent_name, column_name, str(agent_id))
                    self._write_data_frame(agent_data, identifier)
            else:
                identifier = self._get_identifier(agent_name, column_name)
                self._write_data_frame(column_data, identifier)

    def write_time_series_to_disk(self, timeseries_manager: TimeSeriesManager) -> None:
        """Writes time_series of given `timeseries_manager` to disk"""
        for _, name, data in timeseries_manager.get_all_series():
            if data is not None:
                target_path = Path(self._output_folder, name)
                ensure_path_exists(target_path.parent)
                # noinspection PyTypeChecker
                data.to_csv(path_or_buf=target_path, sep=";", header=None, index=None)

    @staticmethod
    def _get_identifier(agent_name: str, column_name: str, agent_id: str = None) -> str:
        """Returns unique identifier for given `agent_name` and (optional) `agent_id` and `column_name`"""
        identifier = str(agent_name)
        if column_name:
            identifier += f"_{column_name}"
        if agent_id:
            identifier += f"_{agent_id}"
        return identifier

    def _write_data_frame(self, data: pd.DataFrame, identifier: str) -> None:
        """
        Appends `data` to existing csv file derived from `identifier` without headers,
        or writes new file with headers instead
        """
        if self._has_file(identifier):
            outfile_name = self._get_outfile_name(identifier)
            data.to_csv(outfile_name, sep=";", index=True, header=False, mode="a")
        else:
            outfile_name = self._create_outfile_name(identifier)
            self._save_outfile_name(outfile_name, identifier)
            data.to_csv(outfile_name, sep=";", index=True, header=True)

    def _has_file(self, identifier: str) -> bool:
        """Returns True if a file for given `identifier` was already written"""
        return identifier in self._files

    def _get_outfile_name(self, identifier: str) -> str:
        """Returns file name for given `agent_name` and (optional) `agent_id`"""
        return self._files[identifier]

    def _create_outfile_name(self, identifier: str) -> Path:
        """Returns fully qualified file name based on given `agent_name` and (optional) `agent_id`"""
        return Path(self._output_folder, f"{identifier}.csv")

    def _save_outfile_name(self, outfile_name: Path, identifier: str) -> None:
        """Stores given name for given `agent_name` and (optional) `agent_id`"""
        self._files[identifier] = outfile_name



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import math
from typing import Dict, Optional

import pandas as pd

from fameio.source import FameTime
from fameio.source.cli.options import TimeOptions, MergingOptions
from fameio.source.logs import log_error_and_raise, log
from fameio.source.time import ConversionException

_ERR_UNIMPLEMENTED = "Time conversion mode '{}' not implemented."


def apply_time_merging(data: Dict[Optional[str], pd.DataFrame], config: Optional[Dict[MergingOptions, int]]) -> None:
    """
    Applies merging of TimeSteps inplace for given `data`

    Args:
        data: one or multiple DataFrames of time series; column `TimeStep` might be modified
        config: dict of MergingOptions defining how to merge the TimeSteps

    Returns:
        Nothing - data is modified inplace
    """
    if config:
        log().debug(f"Grouping TimeSteps...")
        offset = config[MergingOptions.STEPS_BEFORE]
        period = config[MergingOptions.STEPS_AFTER] + config[MergingOptions.STEPS_BEFORE] + 1
        first_positive_focal_point = config[MergingOptions.FOCAL_POINT] % period
        for key in data.keys():
            df = data[key]
            index_columns = df.index.names
            df.reset_index(inplace=True)
            df["TimeStep"] = df["TimeStep"].apply(lambda t: merge_time(t, first_positive_focal_point, offset, period))
            data[key] = df.groupby(by=index_columns).sum()


def merge_time(time_step: int, focal_time: int, offset: int, period: int) -> int:
    """
    Returns `time_step` rounded to its corresponding focal point

    Args:
        time_step: TimeStep to round
        focal_time: First positive focal point
        offset: Range of TimeSteps left of the focal point
        period: Total range of TimeSteps belonging to the focal point

    Returns:
        Corresponding focal point
    """
    return math.floor((time_step + offset - focal_time) / period) * period + focal_time


def apply_time_option(data: Dict[Optional[str], pd.DataFrame], mode: TimeOptions) -> None:
    """
    Applies time option based on given `mode` inplace of given `data`

    Args:
        data: one or multiple DataFrames of time series; column `TimeStep` might be modified (depending on mode)
        mode: name of time conversion mode (derived from Enum)

    Returns:
        Nothing - data is modified inplace
    """
    if mode == TimeOptions.INT:
        log().debug("No time conversion...")
    elif mode == TimeOptions.UTC:
        _convert_time_index(data, "%Y-%m-%d %H:%M:%S")
    elif mode == TimeOptions.FAME:
        _convert_time_index(data, "%Y-%m-%d_%H:%M:%S")
    else:
        log_error_and_raise(ConversionException(_ERR_UNIMPLEMENTED.format(mode)))


def _convert_time_index(data: Dict[Optional[str], pd.DataFrame], datetime_format: str) -> None:
    """
    Inplace replacement of `TimeStep` column in MultiIndex of each item of `data` from FAME's time steps` to DateTime
    in given `date_format`

    Args:
        data: one or multiple DataFrames of time series; column `TimeStep` will be modified
        datetime_format: used for the conversion

    Returns:
        Nothing - data is modified inplace
    """
    log().debug(f"Converting TimeStep to format '{datetime_format}'...")
    for _, df in data.items():
        index_columns = df.index.names
        df.reset_index(inplace=True)
        df["TimeStep"] = df["TimeStep"].apply(lambda t: FameTime.convert_fame_time_step_to_datetime(t, datetime_format))
        df.set_index(keys=index_columns, inplace=True)



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import ast
from typing import Any, Dict, Optional

from fameio.source.scenario.attribute import Attribute
from fameio.source.scenario.exception import (
    assert_or_raise,
    get_or_default,
    get_or_raise,
)
from fameio.source.tools import keys_to_lower


class Agent:
    """Contains specifications for an agent in a scenario"""

    _KEY_TYPE = "Type".lower()
    _KEY_ID = "Id".lower()
    _KEY_ATTRIBUTES = "Attributes".lower()
    _KEY_METADATA = "MetaData".lower()

    _ERR_MISSING_KEY = "Agent requires `key` '{}' but is missing it."
    _ERR_MISSING_TYPE = "Agent requires `type` but is missing it."
    _ERR_MISSING_ID = "Agent requires a positive integer `id` but was '{}'."
    _ERR_DOUBLE_ATTRIBUTE = "Cannot add attribute '{}' to agent {} because it already exists."
    _ERR_ATTRIBUTE_OVERWRITE = "Agent's attributes are already set and would be overwritten."

    def __init__(self, agent_id: int, type_name: str, meta_data: Optional[Dict] = None) -> None:
        """Constructs a new Agent"""
        assert_or_raise(type(agent_id) is int and agent_id >= 0, self._ERR_MISSING_ID.format(agent_id))
        assert_or_raise(bool(type_name and type_name.strip()), self._ERR_MISSING_TYPE)
        self._id: int = agent_id
        self._type_name: str = type_name.strip()
        self._attributes: Dict = {}
        self._meta_data: Optional[Dict] = meta_data if meta_data else {}

    @classmethod
    def from_dict(cls, definitions: dict) -> Agent:
        """Parses an agent from provided `definitions`"""
        definitions = keys_to_lower(definitions)
        agent_type = get_or_raise(definitions, Agent._KEY_TYPE, Agent._ERR_MISSING_TYPE)
        agent_id = get_or_raise(definitions, Agent._KEY_ID, Agent._ERR_MISSING_ID)
        agent = cls(agent_id, agent_type)
        attribute_definitions = get_or_default(definitions, Agent._KEY_ATTRIBUTES, {})
        agent.init_attributes_from_dict(attribute_definitions)
        agent._meta_data = get_or_default(definitions, Agent._KEY_METADATA, {})
        return agent

    def init_attributes_from_dict(self, attributes: Dict[str, Any]) -> None:
        """Initialize Agent `attributes` from dict; Must only be called when creating a new Agent"""
        assert_or_raise(not self._attributes, self._ERR_ATTRIBUTE_OVERWRITE)
        self._attributes = {}
        for name, value in attributes.items():
            full_name = f"{self.type_name}({self.id}): {name}"
            self.add_attribute(name, Attribute(full_name, value))

    def add_attribute(self, name: str, value: Attribute) -> None:
        """Adds a new attribute to the Agent (raise an error if it already exists)"""
        if name in self._attributes:
            raise ValueError(self._ERR_DOUBLE_ATTRIBUTE.format(name, self.display_id))
        self._attributes[name] = value
        self._notify_data_changed()

    def to_dict(self) -> dict:
        """Serializes the Agent content to a dict"""
        result = {Agent._KEY_TYPE: self.type_name, Agent._KEY_ID: self.id}

        if self.attributes:
            attributes_dict = {}
            for attr_name, attr_value in self.attributes.items():
                attributes_dict[attr_name] = attr_value.generic_content
            result[self._KEY_ATTRIBUTES] = attributes_dict
        if self.meta_data:
            result[self._KEY_METADATA] = self.meta_data
        return result

    def to_string(self) -> str:
        """Serializes this agent to a string"""
        return repr(self.to_dict())

    @classmethod
    def from_string(cls, definitions: str) -> Agent:
        return cls.from_dict(ast.literal_eval(definitions))

    def _notify_data_changed(self):
        """Placeholder method used to signal data changes to derived types"""
        pass

    @property
    def id(self) -> int:
        """Returns the ID of the Agent"""
        return self._id

    @property
    def display_id(self) -> str:
        """Returns the ID of the Agent as a string for display purposes"""
        return "#{}".format(self._id)

    @property
    def type_name(self) -> str:
        """Returns the name of the Agent type"""
        return self._type_name

    @property
    def attributes(self) -> Dict[str, Attribute]:
        """Returns dictionary of all Attributes of this agent"""
        return self._attributes

    @property
    def meta_data(self) -> dict:
        """Returns dictionary of all MetaData of this agent"""
        return self._meta_data



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from .path_resolver import PathResolver
from .time import FameTime
from .validator import SchemaValidator
from .writer import ProtoWriter, ProtoWriterException



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Any, Dict, List, Optional

from fameio.source.logs import log
from fameio.source.scenario.attribute import Attribute
from fameio.source.scenario.exception import get_or_default, get_or_raise, log_and_raise
from fameio.source.time import FameTime
from fameio.source.tools import ensure_is_list, keys_to_lower


class Contract:
    """Contract between two Agents of a scenario"""

    _KEY_SENDER = "SenderId".lower()
    _KEY_RECEIVER = "ReceiverId".lower()
    _KEY_PRODUCT = "ProductName".lower()
    _KEY_FIRST_DELIVERY = "FirstDeliveryTime".lower()
    _KEY_INTERVAL = "DeliveryIntervalInSteps".lower()
    _KEY_EXPIRE = "ExpirationTime".lower()
    _KEY_METADATA = "MetaData".lower()
    _KEY_ATTRIBUTES = "Attributes".lower()

    _ERR_MISSING_KEY = "Contract requires key '{}' but is missing it."
    _ERR_MULTI_CONTRACT_CORRUPT = (
        "Definition of Contracts is valid only for One-to-One, One-to-many, Many-to-one, "
        "or N-to-N sender-to-receiver numbers. Found M-to-N pairing in Contract with "
        "Senders: {} and Receivers: {}."
    )
    _ERR_INTERVAL_NOT_POSITIVE = "Contract delivery interval must be a positive integer but was: {}"
    _ERR_SENDER_IS_RECEIVER = "Contract sender and receiver have the same id: {}"
    _ERR_DOUBLE_ATTRIBUTE = "Cannot add attribute '{}' to contract because it already exists."

    def __init__(
        self,
        sender_id: int,
        receiver_id: int,
        product_name: str,
        delivery_interval: int,
        first_delivery_time: int,
        expiration_time: Optional[int] = None,
        meta_data: Optional[dict] = None,
    ) -> None:
        """Constructs a new Contract"""
        assert product_name != ""
        if sender_id == receiver_id:
            log().warning(self._ERR_SENDER_IS_RECEIVER.format(sender_id))
        if delivery_interval <= 0:
            raise ValueError(self._ERR_INTERVAL_NOT_POSITIVE.format(delivery_interval))
        self._sender_id = sender_id
        self._receiver_id = receiver_id
        self._product_name = product_name
        self._delivery_interval = delivery_interval
        self._first_delivery_time = first_delivery_time
        self._expiration_time = expiration_time
        self._meta_data = meta_data
        self._attributes = {}

    def _notify_data_changed(self):
        """Placeholder method used to signal data changes to derived types"""
        pass

    @property
    def product_name(self) -> str:
        """Returns the product name of the contract"""
        return self._product_name

    @property
    def sender_id(self) -> int:
        """Returns the sender ID of the contract"""
        return self._sender_id

    @property
    def display_sender_id(self) -> str:
        """Returns the sender ID of the contract as a string for display purposes"""
        return "#{}".format(self._sender_id)

    @property
    def receiver_id(self) -> int:
        """Returns the receiver ID of the contract"""
        return self._receiver_id

    @property
    def display_receiver_id(self) -> str:
        """Returns the receiver ID of the contract as a string for display purposes"""
        return "#{}".format(self._receiver_id)

    @property
    def delivery_interval(self) -> int:
        """Returns the delivery interval of the contract (in steps)"""
        return self._delivery_interval

    @property
    def first_delivery_time(self) -> int:
        """Returns the first delivery time of the contract"""
        return self._first_delivery_time

    @property
    def expiration_time(self) -> Optional[int]:
        """Returns the expiration time of the contract if available, None otherwise"""
        return self._expiration_time

    @property
    def meta_data(self) -> Optional[dict]:
        """Returns the metadata of the contract if available, None otherwise"""
        return self._meta_data

    @property
    def attributes(self) -> Dict[str, Attribute]:
        """Returns dictionary of all Attributes of the contract"""
        return self._attributes

    def add_attribute(self, name: str, value: Attribute) -> None:
        """Adds a new attribute to the Contract (raise an error if it already exists)"""
        if name in self._attributes:
            raise ValueError(self._ERR_DOUBLE_ATTRIBUTE.format(name))
        self._attributes[name] = value
        self._notify_data_changed()

    @classmethod
    def from_dict(cls, definitions: dict) -> Contract:
        """Parses Contract from given `definitions`"""
        definitions = keys_to_lower(definitions)
        sender_id = get_or_raise(definitions, Contract._KEY_SENDER, Contract._ERR_MISSING_KEY)
        receiver_id = get_or_raise(definitions, Contract._KEY_RECEIVER, Contract._ERR_MISSING_KEY)
        product_name = get_or_raise(definitions, Contract._KEY_PRODUCT, Contract._ERR_MISSING_KEY)
        first_delivery_time = FameTime.convert_string_if_is_datetime(
            get_or_raise(definitions, Contract._KEY_FIRST_DELIVERY, Contract._ERR_MISSING_KEY)
        )
        delivery_interval = get_or_raise(definitions, Contract._KEY_INTERVAL, Contract._ERR_MISSING_KEY)
        expiration_time = get_or_default(definitions, Contract._KEY_EXPIRE, None)
        expiration_time = FameTime.convert_string_if_is_datetime(expiration_time) if expiration_time else None
        meta_data = get_or_default(definitions, Contract._KEY_METADATA, None)
        result = cls(
            sender_id,
            receiver_id,
            product_name,
            delivery_interval,
            first_delivery_time,
            expiration_time,
            meta_data,
        )
        attribute_definitions = get_or_default(definitions, Contract._KEY_ATTRIBUTES, dict())
        result._init_attributes_from_dict(attribute_definitions)
        return result

    def _init_attributes_from_dict(self, attributes: Dict[str, Any]) -> None:
        """Resets Contract `attributes` from dict; Must only be called when creating a new Contract"""
        assert len(self._attributes) == 0
        self._attributes = {}
        for name, value in attributes.items():
            full_name = f"{type}.{id}{name}"
            self.add_attribute(name, Attribute(full_name, value))

    def to_dict(self) -> dict:
        """Serializes the Contract content to a dict"""
        result = {
            self._KEY_SENDER: self.sender_id,
            self._KEY_RECEIVER: self.receiver_id,
            self._KEY_PRODUCT: self.product_name,
            self._KEY_FIRST_DELIVERY: self.first_delivery_time,
            self._KEY_INTERVAL: self.delivery_interval,
        }

        if self.expiration_time is not None:
            result[self._KEY_EXPIRE] = self.expiration_time

        if len(self.attributes) > 0:
            attributes_dict = {}
            for attr_name, attr_value in self.attributes.items():
                attributes_dict[attr_name] = attr_value.generic_content
            result[self._KEY_ATTRIBUTES] = attributes_dict
        if self.meta_data:
            result[self._KEY_METADATA] = self._meta_data

        return result

    @staticmethod
    def split_contract_definitions(multi_definition: dict) -> List[dict]:
        """Splits given `multi_definition` dictionary into list of individual Contract definitions"""
        contracts = []
        base_data = {}
        multi_definition = keys_to_lower(multi_definition)
        for key in [
            Contract._KEY_PRODUCT,
            Contract._KEY_FIRST_DELIVERY,
            Contract._KEY_INTERVAL,
            Contract._KEY_EXPIRE,
            Contract._KEY_METADATA,
            Contract._KEY_ATTRIBUTES,
        ]:
            if key in multi_definition:
                base_data[key] = multi_definition[key]
        senders = ensure_is_list(get_or_raise(multi_definition, Contract._KEY_SENDER, Contract._ERR_MISSING_KEY))
        receivers = ensure_is_list(get_or_raise(multi_definition, Contract._KEY_RECEIVER, Contract._ERR_MISSING_KEY))
        if len(senders) > 1 and len(receivers) == 1:
            for index in range(len(senders)):
                contracts.append(Contract._copy_contract(senders[index], receivers[0], base_data))
        elif len(senders) == 1 and len(receivers) > 1:
            for index in range(len(receivers)):
                contracts.append(Contract._copy_contract(senders[0], receivers[index], base_data))
        elif len(senders) == len(receivers):
            for index in range(len(senders)):
                contracts.append(Contract._copy_contract(senders[index], receivers[index], base_data))
        else:
            log_and_raise(Contract._ERR_MULTI_CONTRACT_CORRUPT.format(senders, receivers))
        return contracts

    @staticmethod
    def _copy_contract(sender: int, receiver: int, base_data: dict) -> dict:
        """Returns a new contract definition dictionary, with given `sender` and `receiver` and copied `base_data`"""
        contract = {
            Contract._KEY_SENDER: sender,
            Contract._KEY_RECEIVER: receiver,
        }
        contract.update(base_data)
        return contract



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Any, Dict, List, Union

from fameprotobuf.DataStorage_pb2 import DataStorage
from fameprotobuf.InputFile_pb2 import InputData
from fameprotobuf.Field_pb2 import NestedField
from fameprotobuf.Contract_pb2 import ProtoContract
from fameprotobuf.Model_pb2 import ModelData

from fameio.source.logs import log_error_and_raise, log
from fameio.source.results.reader import Reader
from fameio.source.scenario import (
    Agent,
    Attribute,
    Contract,
    GeneralProperties,
    Scenario,
)
from fameio.source.schema import Schema
from fameio.source.schema.attribute import AttributeSpecs, AttributeType
from fameio.source.schema.java_packages import JavaPackages
from fameio.source.series import TimeSeriesManager
from fameio.source.time import FameTime
from fameio.source.tools import ensure_is_list


class ProtoWriterException(Exception):
    """Indicates an error during writing of protobuf file"""

    pass


class ProtoWriter:
    """Writes a given scenario to protobuf file"""

    _FAME_PROTOBUF_STREAM_HEADER = "famecoreprotobufstreamfilev001"  # noqa

    _TYPE_NOT_IMPLEMENTED = "AttributeType '{}' not implemented."
    _CONTRACT_UNSUPPORTED = (
        "Unsupported value for Contract Attribute '{}'; "
        "Only support `int`, `float`, `enum` or `dict` types are supported here."
    )
    _USING_DEFAULT = "Using provided Default for Attribute: '{}'."
    _NO_FILE_SPECIFIED = "Could not write to '{}'. Please specify a valid output file."

    _INFO_WRITING = "Writing scenario to protobuf file `{}`"
    _INFO_WRITING_COMPLETED = "Saved protobuf file `{}` to disk"

    def __init__(self, file_path: Path, time_series_manager: TimeSeriesManager) -> None:
        self.file_path: Path = file_path
        self._time_series_manager: TimeSeriesManager = time_series_manager

    def write_validated_scenario(self, scenario: Scenario) -> None:
        """Writes given validated Scenario to file"""
        pb_data_storage = self._create_protobuf_from_scenario(scenario)
        self._write_protobuf_to_disk(pb_data_storage)

    def _create_protobuf_from_scenario(self, scenario: Scenario) -> DataStorage:
        """Returns given `scenario` written to new DataStorage protobuf"""
        log().info("Converting scenario to protobuf.")
        pb_data_storage = DataStorage()
        pb_input = pb_data_storage.input

        self._set_general_properties(pb_input, scenario.general_properties)
        self._add_agents(pb_input, scenario.agents, scenario.schema)
        self._add_contracts(pb_input, scenario.contracts)
        self._set_time_series(pb_input)
        self._set_schema(pb_input, scenario.schema)

        self._set_java_package_names(pb_data_storage.model, scenario.schema.packages)
        return pb_data_storage

    @staticmethod
    def _set_general_properties(pb_input: InputData, general_properties: GeneralProperties) -> None:
        """Saves a scenario's general properties to specified protobuf `pb_input` container"""
        log().info("Adding General Properties")
        pb_input.runId = general_properties.run_id
        pb_input.simulation.startTime = general_properties.simulation_start_time
        pb_input.simulation.stopTime = general_properties.simulation_stop_time
        pb_input.simulation.randomSeed = general_properties.simulation_random_seed
        pb_input.output.interval = general_properties.output_interval
        pb_input.output.process = general_properties.output_process

    def _add_agents(self, pb_input: InputData, agents: List[Agent], schema: Schema) -> None:
        """Triggers setting of `agents` to `pb_input`"""
        log().info("Adding Agents")
        for agent in agents:
            pb_agent = self._set_agent(pb_input.agent.add(), agent)
            attribute_specs = schema.agent_types[agent.type_name].attributes
            self._set_attributes(pb_agent, agent.attributes, attribute_specs)
            pb_agent.metadata = repr(agent.meta_data)

    @staticmethod
    def _set_agent(pb_agent: InputData.AgentDao, agent: Agent) -> InputData.AgentDao:
        """Saves type and id of given `agent` to protobuf `pb_agent` container. Returns given `pb_agent`"""
        pb_agent.className = agent.type_name
        pb_agent.id = agent.id
        return pb_agent

    def _set_attributes(
        self,
        pb_parent: Union[InputData.AgentDao, NestedField],
        attributes: Dict[str, Attribute],
        specs: Dict[str, AttributeSpecs],
    ) -> None:
        """Assigns `attributes` to protobuf fields of given `pb_parent` - cascades for nested Attributes"""
        values_not_set = [key for key in specs.keys()]
        for name, attribute in attributes.items():
            pb_field = self._add_field(pb_parent, name)
            attribute_specs = specs[name]
            values_not_set.remove(name)
            attribute_type = attribute_specs.attr_type
            if attribute_type is AttributeType.BLOCK:
                if attribute_specs.is_list:
                    for index, entry in enumerate(attribute.nested_list):
                        pb_inner = self._add_field(pb_field, str(index))
                        self._set_attributes(pb_inner, entry, attribute_specs.nested_attributes)
                else:
                    self._set_attributes(pb_field, attribute.nested, attribute_specs.nested_attributes)
            else:
                self._set_attribute(pb_field, attribute.value, attribute_type)
        for name in values_not_set:
            attribute_specs = specs[name]
            if attribute_specs.is_mandatory:
                pb_field = self._add_field(pb_parent, name)
                self._set_attribute(pb_field, attribute_specs.default_value, attribute_specs.attr_type)
                log().info(self._USING_DEFAULT.format(name))

    @staticmethod
    def _add_field(pb_parent: Union[InputData.AgentDao, NestedField], name: str) -> NestedField:
        """Returns new field with given `name` that is added to given `pb_parent`"""
        pb_field = pb_parent.field.add()
        pb_field.fieldName = name
        return pb_field

    def _set_attribute(self, pb_field: NestedField, value: Any, attribute_type: AttributeType) -> None:
        """Sets given `value` to given protobuf `pb_field` depending on specified `attribute_type`"""
        if attribute_type is AttributeType.INTEGER:
            pb_field.intValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.DOUBLE:
            pb_field.doubleValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.LONG:
            pb_field.longValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.TIME_STAMP:
            pb_field.longValue.extend(ensure_is_list(FameTime.convert_string_if_is_datetime(value)))
        elif attribute_type in (AttributeType.ENUM, AttributeType.STRING):
            pb_field.stringValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.TIME_SERIES:
            pb_field.seriesId = self._time_series_manager.get_series_id_by_identifier(value)
        else:
            log_error_and_raise(ProtoWriterException(self._TYPE_NOT_IMPLEMENTED.format(attribute_type)))

    @staticmethod
    def _add_contracts(pb_input: InputData, contracts: List[Contract]) -> None:
        """Triggers setting of `contracts` to `pb_input`"""
        log().info("Adding Contracts")
        for contract in contracts:
            pb_contract = ProtoWriter._set_contract(pb_input.contract.add(), contract)
            ProtoWriter._set_contract_attributes(pb_contract, contract.attributes)
            pb_contract.metadata = repr(contract.meta_data)

    @staticmethod
    def _set_contract(pb_contract: ProtoContract, contract: Contract) -> ProtoContract:
        """Saves given `contract` details to protobuf container `pb_contract`. Returns given `pb_contract`"""
        pb_contract.senderId = contract.sender_id
        pb_contract.receiverId = contract.receiver_id
        pb_contract.productName = contract.product_name
        pb_contract.firstDeliveryTime = contract.first_delivery_time
        pb_contract.deliveryIntervalInSteps = contract.delivery_interval
        if contract.expiration_time:
            pb_contract.expirationTime = contract.expiration_time
        return pb_contract

    @staticmethod
    def _set_contract_attributes(
        pb_parent: Union[ProtoContract, NestedField], attributes: Dict[str, Attribute]
    ) -> None:
        """Assign (nested) Attributes to given protobuf container `pb_parent`"""
        for name, attribute in attributes.items():
            log().debug("Assigning contract attribute `{}`.".format(name))
            pb_field = ProtoWriter._add_field(pb_parent, name)

            if attribute.has_value:
                value = attribute.value
                if isinstance(value, int):
                    pb_field.intValue.extend([value])
                elif isinstance(value, float):
                    pb_field.doubleValue.extend([value])
                elif isinstance(value, str):
                    pb_field.stringValue.extend([value])
                else:
                    log_error_and_raise(ProtoWriterException(ProtoWriter._CONTRACT_UNSUPPORTED.format(str(attribute))))
            elif attribute.has_nested:
                ProtoWriter._set_contract_attributes(pb_field, attribute.nested)

    def _set_time_series(self, pb_input: InputData) -> None:
        """Adds all time series from TimeSeriesManager to given `pb_input`"""
        log().info("Adding TimeSeries")
        for unique_id, identifier, data in self._time_series_manager.get_all_series():
            pb_series = pb_input.timeSeries.add()
            pb_series.seriesId = unique_id
            pb_series.seriesName = identifier
            ProtoWriter._add_rows_to_series(pb_series, data)

    @staticmethod
    def _add_rows_to_series(series: InputData.TimeSeriesDao, data_frame) -> None:
        for key, value in data_frame.itertuples(index=False):
            row = series.row.add()
            row.timeStep = int(key)
            row.value = value

    @staticmethod
    def _set_schema(pb_input: InputData, schema: Schema) -> None:
        """Sets the given `schema` `pb_input`"""
        log().info("Adding Schema")
        pb_input.schema = schema.to_string()

    @staticmethod
    def _set_java_package_names(pb_model: ModelData, java_packages: JavaPackages) -> None:
        """Adds given JavaPackages names to given ModelData section"""
        pb_packages = pb_model.packages
        pb_packages.agent.extend(java_packages.agents)
        pb_packages.dataItem.extend(java_packages.data_items)
        pb_packages.portable.extend(java_packages.portables)

    def _write_protobuf_to_disk(self, pb_data_storage: DataStorage) -> None:
        """Writes given `protobuf_input_data` to disk"""
        log().info(self._INFO_WRITING.format(self.file_path))
        try:
            with open(self.file_path, "wb") as file:
                serialised_data_storage = pb_data_storage.SerializeToString()
                file.write(self._FAME_PROTOBUF_STREAM_HEADER.encode(Reader.HEADER_ENCODING))
                file.write(len(serialised_data_storage).to_bytes(Reader.BYTES_DEFINING_MESSAGE_LENGTH, byteorder="big"))
                file.write(serialised_data_storage)
        except OSError as e:
            log_error_and_raise(ProtoWriterException(ProtoWriter._NO_FILE_SPECIFIED.format(self.file_path), e))
        log().info(self._INFO_WRITING_COMPLETED.format(self.file_path))



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import typing
from enum import Enum, auto
from typing import Any, Dict

from fameio.source.logs import log
from fameio.source.schema.exception import SchemaException
from fameio.source.time import FameTime
from fameio.source.tools import keys_to_lower


class AttributeType(Enum):
    """Specifies known types that Attributes can take"""

    INTEGER = auto()
    DOUBLE = auto()
    LONG = auto()
    TIME_STAMP = auto()
    STRING = auto()
    ENUM = auto()
    TIME_SERIES = auto()
    BLOCK = auto()

    def convert_string_to_type(self, value: str):
        """Converts a given string to this AttributeType's data format"""
        if self is AttributeType.INTEGER or self is AttributeType.LONG:
            return int(value)
        elif self is AttributeType.DOUBLE:
            return float(value)
        elif self is AttributeType.TIME_STAMP:
            return FameTime.convert_string_if_is_datetime(value)
        elif self is AttributeType.ENUM or self is AttributeType.STRING:
            return str(value)
        elif self is AttributeType.TIME_SERIES:
            return float(value)
        else:
            raise ValueError("String conversion not supported for '{}'.".format(self))


class AttributeSpecs:
    """Schema Definition of a single Attribute (with possible inner Attributes) of an agent"""

    _MISSING_SPEC_DEFAULT = "Missing '{}' specification for Attribute '{}' - assuming {}."
    _MISSING_TYPE = "'AttributeType' not declare for Attribute '{}'."
    _INVALID_TYPE = "'{}' is not a valid type for an Attribute."
    _DEFAULT_NOT_LIST = "Attribute is list, but provided Default '{}' is not a list."
    _DEFAULT_INCOMPATIBLE = "Default '{}' can not be converted to AttributeType '{}'."
    _DEFAULT_DISALLOWED = "Default '{}' is not an allowed value."
    _LIST_DISALLOWED = "Attribute '{}' of type TIME_SERIES cannot be a list."
    _VALUES_ILL_FORMAT = "Only List and Dictionary is supported for 'Values' but was: {}"

    _KEY_MANDATORY = "Mandatory".lower()
    _KEY_LIST = "List".lower()
    _KEY_TYPE = "AttributeType".lower()
    _KEY_NESTED = "NestedAttributes".lower()
    _KEY_VALUES = "Values".lower()
    _KEY_DEFAULT = "Default".lower()
    _KEY_HELP = "Help".lower()

    def __init__(self, name: str, definition: dict):
        """Loads Attribute from given `definition`"""
        self._full_name = name
        definition = keys_to_lower(definition)

        if AttributeSpecs._KEY_MANDATORY in definition:
            self._is_mandatory = definition[AttributeSpecs._KEY_MANDATORY]
        else:
            self._is_mandatory = True
            log().warning(AttributeSpecs._MISSING_SPEC_DEFAULT.format(AttributeSpecs._KEY_MANDATORY, name, True))

        if AttributeSpecs._KEY_LIST in definition:
            self._is_list = definition[AttributeSpecs._KEY_LIST]
        else:
            self._is_list = False
            log().warning(AttributeSpecs._MISSING_SPEC_DEFAULT.format(AttributeSpecs._KEY_LIST, name, False))

        if AttributeSpecs._KEY_TYPE in definition:
            self._attr_type = AttributeSpecs._get_type_for_name(definition[AttributeSpecs._KEY_TYPE])
        else:
            log().error(AttributeSpecs._MISSING_TYPE.format(name))
            raise SchemaException(AttributeSpecs._MISSING_TYPE.format(name))

        if self._attr_type == AttributeType.TIME_SERIES and self._is_list:
            raise SchemaException(AttributeSpecs._LIST_DISALLOWED.format(name))

        self._values = []
        if AttributeSpecs._KEY_VALUES in definition:
            value_definition = definition[AttributeSpecs._KEY_VALUES]
            if value_definition:
                self._values = self._read_values(value_definition)

        self._default_value = None
        if AttributeSpecs._KEY_DEFAULT in definition:
            provided_value = definition[AttributeSpecs._KEY_DEFAULT]
            if self._is_list:
                self._default_value = self._convert_list(provided_value)
            else:
                self._default_value = self._convert_to_data_type(provided_value)

        self._nested_attributes = {}
        if AttributeSpecs._KEY_NESTED in definition:
            for nested_name, nested_details in definition[AttributeSpecs._KEY_NESTED].items():
                full_name = name + "." + nested_name
                self._nested_attributes[nested_name] = AttributeSpecs(full_name, nested_details)

        self._help = None
        if AttributeSpecs._KEY_HELP in definition:
            self._help = definition[AttributeSpecs._KEY_HELP]

    @staticmethod
    def _read_values(definition: Any) -> list:
        """Returns list of values extracted from given `definition` if it is a list or dict"""
        if isinstance(definition, list):
            return definition
        elif isinstance(definition, dict):
            return [key for key in definition.keys()]
        else:
            raise SchemaException(AttributeSpecs._VALUES_ILL_FORMAT.format(definition))

    def _convert_list(self, values) -> list:
        """Converts all entries in given `values` list to this attribute data type and returns this new list"""
        if isinstance(values, list):
            return [self._convert_to_data_type(item) for item in values]
        else:
            raise SchemaException(AttributeSpecs._DEFAULT_NOT_LIST.format(values))

    def _convert_to_data_type(self, value: str):
        """Converts a given single `value` to this Attribute's data type"""
        try:
            converted = self._attr_type.convert_string_to_type(value)
            if self._values and converted not in self._values:
                raise SchemaException(AttributeSpecs._DEFAULT_DISALLOWED.format(value))
            return converted
        except ValueError:
            raise SchemaException(AttributeSpecs._DEFAULT_INCOMPATIBLE.format(value, self._attr_type))

    @property
    def attr_type(self) -> AttributeType:
        """Returns AttributeType of this attribute"""
        return self._attr_type

    @property
    def values(self) -> list:
        """Returns the list of allowed values for this attribute"""
        return self._values

    @property
    def is_list(self) -> bool:
        """Return True if this attribute type is a list"""
        return self._is_list

    @property
    def has_nested_attributes(self) -> bool:
        """Returns True if nested attributes are defined"""
        return bool(self._nested_attributes)

    @property
    def nested_attributes(self) -> Dict[str, "AttributeSpecs"]:
        """Returns list of nested Attributes of this Attribute or an empty dict if no nested attributes are defined"""
        return self._nested_attributes

    @property
    def has_default_value(self) -> bool:
        """Return True if a default value is available"""
        return self._default_value is not None

    @property
    def default_value(self) -> typing.Optional[typing.Any]:
        """Return the default value of this attribute, if any"""
        return self._default_value

    @property
    def is_mandatory(self) -> bool:
        """Return True if this attribute is mandatory"""
        return self._is_mandatory

    @property
    def full_name(self) -> str:
        """Returns name including name of enclosing parent attributes"""
        return self._full_name

    @staticmethod
    def _get_type_for_name(name: str) -> AttributeType:
        """Returns the AttributeType matching the given `name` converted to upper case"""
        try:
            return AttributeType[name.upper()]
        except KeyError:
            raise SchemaException(AttributeSpecs._INVALID_TYPE.format(name))

    @property
    def has_help_text(self) -> bool:
        """Return True if a help_text is available"""
        return bool(self._help)

    @property
    def help_text(self) -> str:
        """Return the help_text of this attribute, if any, otherwise an empty string"""
        return self._help if self.has_help_text else ""



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from fameio.source.logs import log
from fameio.source.scenario.exception import get_or_default, get_or_raise
from fameio.source.time import FameTime
from fameio.source.tools import keys_to_lower


class GeneralProperties:
    """Hosts general properties of a scenario"""

    _KEY_RUN = "RunId".lower()
    _KEY_SIMULATION = "Simulation".lower()
    _KEY_START = "StartTime".lower()
    _KEY_STOP = "StopTime".lower()
    _KEY_SEED = "RandomSeed".lower()
    _KEY_OUTPUT = "Output".lower()
    _KEY_INTERVAL = "Interval".lower()
    _KEY_PROCESS = "Process".lower()

    _MISSING_KEY = "General Properties requires key '{}' but it is missing."
    _SIMULATION_DURATION = "Simulation starts after its end time - check start and stop times."

    def __init__(
        self,
        run_id: int,
        simulation_start_time: int,
        simulation_stop_time: int,
        simulation_random_seed: int,
        output_interval: int,
        output_process: int,
    ) -> None:
        if simulation_stop_time < simulation_start_time:
            log().warning(GeneralProperties._SIMULATION_DURATION)
        self._run_id = run_id
        self._simulation_start_time = simulation_start_time
        self._simulation_stop_time = simulation_stop_time
        self._simulation_random_seed = simulation_random_seed
        self._output_interval = output_interval
        self._output_process = output_process

    @classmethod
    def from_dict(cls, definitions: dict) -> GeneralProperties:
        """Parse general properties from provided `definitions`"""
        definitions = keys_to_lower(definitions)
        run_id = get_or_default(definitions, GeneralProperties._KEY_RUN, 1)

        simulation_definition = keys_to_lower(
            get_or_raise(
                definitions,
                GeneralProperties._KEY_SIMULATION,
                GeneralProperties._MISSING_KEY,
            )
        )
        start_time = FameTime.convert_string_if_is_datetime(
            get_or_raise(
                simulation_definition,
                GeneralProperties._KEY_START,
                GeneralProperties._MISSING_KEY,
            )
        )
        stop_time = FameTime.convert_string_if_is_datetime(
            get_or_raise(
                simulation_definition,
                GeneralProperties._KEY_STOP,
                GeneralProperties._MISSING_KEY,
            )
        )
        random_seed = get_or_default(simulation_definition, GeneralProperties._KEY_SEED, 1)

        output_definitions = keys_to_lower(get_or_default(definitions, GeneralProperties._KEY_OUTPUT, dict()))
        output_interval = get_or_default(output_definitions, GeneralProperties._KEY_INTERVAL, 100)
        output_process = get_or_default(output_definitions, GeneralProperties._KEY_PROCESS, 0)

        return cls(run_id, start_time, stop_time, random_seed, output_interval, output_process)

    def to_dict(self) -> dict:
        """Serializes the general properties to a dict"""
        result = {self._KEY_RUN: self._run_id}
        simulation_dict = {
            self._KEY_START: self.simulation_start_time,
            self._KEY_STOP: self.simulation_stop_time,
            self._KEY_SEED: self.simulation_random_seed,
        }
        result[self._KEY_SIMULATION] = simulation_dict
        output_dict = {self._KEY_INTERVAL: self.output_interval, self._KEY_PROCESS: self.output_process}
        result[self._KEY_OUTPUT] = output_dict
        return result

    @property
    def run_id(self) -> int:
        """Returns the run ID"""
        return self._run_id

    @property
    def simulation_start_time(self) -> int:
        """Returns the simulation start time"""
        return self._simulation_start_time

    @property
    def simulation_stop_time(self) -> int:
        """Returns the simulation stop time"""
        return self._simulation_stop_time

    @property
    def simulation_random_seed(self) -> int:
        """Returns the simulation random seed"""
        return self._simulation_random_seed

    @property
    def output_interval(self) -> int:
        """Returns the output interval"""
        return self._output_interval

    @property
    def output_process(self) -> int:
        """Returns the output process"""
        return self._output_process



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import argparse
from pathlib import Path
from typing import List, Dict, Any, Optional

from fameio.source.cli.options import Options
from fameio.source.cli.parser import (
    add_file_argument,
    add_log_level_argument,
    add_logfile_argument,
    add_output_argument,
    map_namespace_to_options_dict,
)

CLI_DEFAULTS = {
    Options.FILE: None,
    Options.LOG_LEVEL: "WARN",
    Options.LOG_FILE: None,
    Options.OUTPUT: Path("config.pb"),
}

_INFILE_PATH_HELP = "provide path to configuration file"
_OUTFILE_PATH_HELP = "provide file-path for the file to generate"


def handle_args(args: List[str], defaults: Optional[Dict[Options, Any]] = None) -> Dict[Options, Any]:
    """
    Converts given `arguments` and returns a configuration for the make_config script

    Args:
        args: list of (command line) arguments, e.g., ['-f', 'my_file']; arg values take precedence over defaults
        defaults: optional default values used for unspecified parameters; missing defaults are replaced by CLI defaults

    Returns:
        final configuration compiled from (given) `defaults` and given `args`
    """
    parser = _prepare_parser(defaults)
    parsed = parser.parse_args(args)
    return map_namespace_to_options_dict(parsed)


def _prepare_parser(defaults: Optional[Dict[Options, Any]]) -> argparse.ArgumentParser:
    """
    Creates a parser with given defaults to handle `make_config` configuration arguments

    Returns:
        new parser using given defaults for its arguments; if a default is not specified, hard-coded defaults are used
    """
    defaults = defaults if (defaults is not None) else {}
    parser = argparse.ArgumentParser()
    add_file_argument(parser, _get_default(defaults, Options.FILE), _INFILE_PATH_HELP)
    add_log_level_argument(parser, _get_default(defaults, Options.LOG_LEVEL))
    add_logfile_argument(parser, _get_default(defaults, Options.LOG_FILE))
    add_output_argument(parser, _get_default(defaults, Options.OUTPUT), _OUTFILE_PATH_HELP)
    return parser


def _get_default(defaults: dict, option: Options) -> Any:
    """Returns default for given `option` or, if missing, its cli default"""
    return defaults.get(option, CLI_DEFAULTS[option])



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from typing import List, Dict, Iterable, Optional

import pandas as pd
from fameprotobuf.DataStorage_pb2 import DataStorage
from fameprotobuf.Services_pb2 import Output

from fameio.source.results.agent_type import AgentTypeLog
from fameio.source.results.data_transformer import DataTransformer


class OutputDAO:
    """Grants convenient access to content of Output protobuf messages for given DataStorages"""

    def __init__(self, data_storages: List[DataStorage], agent_type_log: AgentTypeLog) -> None:
        self._agent_type_log = agent_type_log
        outputs = self._extract_output_from_data_storages(data_storages)
        self._agent_type_log.update_agents(self._extract_new_agent_types(outputs))
        self._all_series = self._extract_series(outputs)

    @staticmethod
    def _extract_output_from_data_storages(data_storages: List[DataStorage]) -> List[Output]:
        """Returns list of Outputs extracted from given `data_storages`"""
        if data_storages is None:
            data_storages = []
        return [data_storage.output for data_storage in data_storages if data_storage.HasField("output")]

    @staticmethod
    def _extract_new_agent_types(outputs: List[Output]) -> Dict[str, Output.AgentType]:
        """Returns dict of agent names mapped to its type defined in given `outputs`"""
        list_of_agent_type_lists = [output.agentType for output in outputs if len(output.agentType) > 0]
        list_of_agent_types = [item for sublist in list_of_agent_type_lists for item in sublist]
        return {item.className: item for item in list_of_agent_types}

    @staticmethod
    def _extract_series(outputs: List[Output]) -> Dict[str, List[Output.Series]]:
        """Returns series data from associated `outputs` mapped to the className of its agent"""
        list_of_series_lists = [output.series for output in outputs if len(output.series) > 0]
        list_of_series = [series for sublist in list_of_series_lists for series in sublist]

        series_per_class_name = {}
        for series in list_of_series:
            if series.className not in series_per_class_name:
                series_per_class_name[series.className] = []
            series_per_class_name[series.className].append(series)
        return series_per_class_name

    def get_sorted_agents_to_extract(self) -> Iterable[str]:
        """Returns iterator of requested and available agent names in ascending order by count of series"""
        all_series = self._get_agent_names_by_series_count_ascending()
        filtered_series = [agent_name for agent_name in all_series if self._agent_type_log.is_requested(agent_name)]
        return iter(filtered_series)

    def _get_agent_names_by_series_count_ascending(self) -> List[str]:
        """Returns list of agent type names sorted by their amount of series"""
        length_per_agent_types = {agent_name: len(value) for agent_name, value in self._all_series.items()}
        sorted_dict = sorted(length_per_agent_types.items(), key=lambda item: item[1])
        return [agent_name for agent_name, _ in sorted_dict]

    def get_agent_data(self, agent_name: str, data_transformer: DataTransformer) -> Dict[Optional[str], pd.DataFrame]:
        """
        Returns DataFrame(s) containing all data of given `agent` - data is removed after the first call
        Depending on the chosen ResolveOption the dict contains one DataFrame for the simple (and merged columns),
        or, in `SPLIT` mode, additional DataFrames mapped to each complex column's name.
        """
        agent_series = self._all_series.pop(agent_name) if agent_name in self._all_series else []
        agent_type = self._agent_type_log.get_agent_type(agent_name)
        return data_transformer.extract_agent_data(agent_series, agent_type)



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import struct
import typing
from abc import ABC, abstractmethod
from typing import IO, List

from fameprotobuf.DataStorage_pb2 import DataStorage
from google.protobuf.message import DecodeError

from fameio.source.logs import log


class Reader(ABC):
    """Abstract base class for protobuf file readers"""

    _WARN_NO_HEADER = "No header recognised in file. File might be deprecated or corrupted."  # noqa
    _ERR_FILE_CORRUPT_NEGATIVE_LENGTH = "Corrupt file, message length must be positive."
    _ERR_FILE_CORRUPT_MISSING_DATA = "Trying to read corrupt file caused by inconsistent message length."
    _ERR_UNSUPPORTED_MODE = "Ignoring memory saving mode: not supported for files created with `fame-core<1.4`."
    _ERR_PARSING_FAILED = "File Corrupt. Could not parse file content."
    _DEBUG_FILE_END_REACHED = "Reached expected end of file."

    _HEADER_LENGTH = 30
    HEADER_ENCODING = "utf-8"
    BYTES_DEFINING_MESSAGE_LENGTH = 4
    _READER_HEADERS = {
        "famecoreprotobufstreamfilev001": lambda file, mode: ReaderV1(file, mode),  # noqa
    }

    def __init__(self, file: IO, read_single) -> None:
        self._file = file
        self._read_single = read_single

    @abstractmethod
    def read(self) -> List[DataStorage]:
        """Reads associated filestream and returns one or multiple DataStorage(s) or empty list"""

    @staticmethod
    def get_reader(file: IO, read_single: bool = False) -> Reader:
        """
        Returns reader matching the given file header

        Args:
            file: to be read by the returned Reader
            read_single: if True, the returned Reader's `read()` method gets one messages at a time

        Returns:
            Reader that can read the specified file
        """
        log().debug("Reading file headers...")
        try:
            header = file.read(Reader._HEADER_LENGTH).decode(Reader.HEADER_ENCODING)
            return Reader._READER_HEADERS[header](file, read_single)
        except (KeyError, UnicodeDecodeError):
            log().warning(Reader._WARN_NO_HEADER)
            file.seek(0)
            if read_single:
                log().error(Reader._ERR_UNSUPPORTED_MODE)
            return ReaderV0(file, False)

    @typing.final
    def _read_message_length(self) -> int:
        """Returns length of next DataStorage message in file"""
        message_length_byte = self._file.read(self.BYTES_DEFINING_MESSAGE_LENGTH)
        if not message_length_byte:
            log().debug(self._DEBUG_FILE_END_REACHED)
            message_length_int = 0
        else:
            message_length_int = struct.unpack(">i", message_length_byte)[0]
        return message_length_int

    @typing.final
    def _read_data_storage_message(self, message_length: int = None) -> DataStorage:
        """
        Returns given `data_storage` read from current file position and following `message_length` bytes.
        If `message_length` is omitted, the rest of the file is read. If no message is found, None is returned.
        """
        if message_length is None:
            message = self._file.read()
        elif message_length > 0:
            message = self._file.read(message_length)
        else:
            raise IOError(self._ERR_FILE_CORRUPT_NEGATIVE_LENGTH)
        if message_length and len(message) != message_length:
            log().error(self._ERR_FILE_CORRUPT_MISSING_DATA)
        return self._parse_to_data_storage(message) if message else None

    @staticmethod
    @typing.final
    def _parse_to_data_storage(message: bytes) -> DataStorage:
        data_storage = DataStorage()
        try:
            data_storage.ParseFromString(message)
        except DecodeError:
            raise IOError(Reader._ERR_PARSING_FAILED)
        return data_storage


class ReaderV0(Reader):
    """Reader class for deprecated `fame-core<1.4` output without any header"""

    _WARN_DEPRECATED = "DeprecationWarning: Please consider updating to `FAME-Core>=1.4` and `fameio>=1.6`"

    def __init__(self, file: IO, read_single):
        super().__init__(file, read_single)
        log().warning(self._WARN_DEPRECATED)

    def read(self) -> List[DataStorage]:
        result = self._read_data_storage_message()
        return [result] if result else []


class ReaderV1(Reader):
    """Reader class for `fame-core>=1.4` output with header of version v001"""

    def read(self) -> List[DataStorage]:
        messages = []
        while True:
            message_length = self._read_message_length()
            if message_length == 0:
                break
            messages.append(self._read_data_storage_message(message_length))
            if self._read_single:
                break
        log().debug(f"Read {len(messages)} messages from file.")
        return messages



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from fameio.source.schema.schema import Schema
from fameio.source.scenario.generalproperties import GeneralProperties
from fameio.source.scenario.agent import Agent
from fameio.source.scenario.contract import Contract


class FameIOFactory:
    """Factory used to instantiate the types defined in a scenario file.
    This allows a client to subclass some types in order to extend what a scenario can contain.
    """

    @staticmethod
    def new_schema_from_dict(definitions: dict) -> Schema:
        """Constructs a new Schema from provided `definitions`"""
        return Schema.from_dict(definitions)

    @staticmethod
    def new_general_properties_from_dict(definitions: dict) -> GeneralProperties:
        """Constructs a new GeneralProperties instance from provided `definitions`"""
        return GeneralProperties.from_dict(definitions)

    @staticmethod
    def new_agent_from_dict(definitions: dict) -> Agent:
        """Constructs a new Agent from provided `definitions`"""
        return Agent.from_dict(definitions)

    @staticmethod
    def new_contract_from_dict(definitions: dict) -> Contract:
        """Constructs a new Contract from provided `definitions`"""
        return Contract.from_dict(definitions)



#!/usr/bin/env python
import sys

from fameio.scripts.convert_results import DEFAULT_CONFIG as DEFAULT_CONVERT_CONFIG
from fameio.scripts.convert_results import run as convert_results
from fameio.scripts.make_config import run as make_config
from fameio.source.cli.convert_results import handle_args as handle_convert_results_args
from fameio.source.cli.make_config import handle_args as handle_make_config_args


# noinspection PyPep8Naming
def makeFameRunConfig():
    run_config = handle_make_config_args(sys.argv[1:])
    make_config(run_config)


# noinspection PyPep8Naming
def convertFameResults():
    run_config = handle_convert_results_args(sys.argv[1:], DEFAULT_CONVERT_CONFIG)
    convert_results(run_config)



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Dict, List, Any

from fameio.source.logs import log_error_and_raise, log
from fameio.source.schema.exception import SchemaException
from fameio.source.schema.attribute import AttributeSpecs
from fameio.source.tools import keys_to_lower


class AgentType:
    """Schema definitions for an Agent type"""

    _ERR_NAME_INVALID = "'{}' is not a valid name for AgentTypes"
    _ERR_PRODUCTS_NO_STRING = "Product definition of AgentType '{}' contains item(s) / key(s) other than string: '{}'"
    _ERR_PRODUCTS_UNKNOWN_STRUCTURE = "Product definition of AgentType '{}' is neither list nor dictionary: '{}'"

    _NO_ATTRIBUTES = "Agent '{}' has no specified 'Attributes'."
    _NO_PRODUCTS = "Agent '{}' has no specified 'Products'."
    _NO_OUTPUTS = "Agent '{}' has no specified 'Outputs'."
    _NO_METADATA = "Agent '{}' has no specified 'Metadata'."

    _KEY_ATTRIBUTES = "Attributes".lower()
    _KEY_PRODUCTS = "Products".lower()
    _KEY_OUTPUTS = "Outputs".lower()
    _KEY_METADATA = "MetaData".lower()

    def __init__(self, name: str):
        """
        Initialise a new AgentType

        Args:
            name: name of the AgenType - must not be None or empty
        """
        if not name or name.isspace():
            log_error_and_raise(SchemaException(AgentType._ERR_NAME_INVALID.format(name)))
        self._name = name
        self._attributes = {}
        self._products = {}
        self._outputs = {}
        self._metadata = {}

    @classmethod
    def from_dict(cls, name: str, definitions: dict) -> AgentType:
        """
        Creates AgentType with given `name` from specified dictionary

        Args:
            name: of the agent type
            definitions: of the agent type specifying, e.g., its attributes and products

        Returns:
            a new instance of AgentType
        """
        agent_type = cls(name)

        definition = keys_to_lower(definitions)
        if AgentType._KEY_ATTRIBUTES in definition:
            for attribute_name, attribute_details in definition[AgentType._KEY_ATTRIBUTES].items():
                full_name = name + "." + attribute_name
                agent_type._attributes[attribute_name] = AttributeSpecs(full_name, attribute_details)
        else:
            log().info(AgentType._NO_ATTRIBUTES.format(name))

        if AgentType._KEY_PRODUCTS in definition and definition[AgentType._KEY_PRODUCTS]:
            agent_type._products.update(AgentType._read_products(definition[AgentType._KEY_PRODUCTS], name))
        else:
            log().info(AgentType._NO_PRODUCTS.format(name))

        if AgentType._KEY_OUTPUTS in definition:
            outputs_to_add = definition[AgentType._KEY_OUTPUTS]
            if outputs_to_add:
                agent_type._outputs.update(outputs_to_add)
        else:
            log().debug(AgentType._NO_OUTPUTS.format(name))

        if AgentType._KEY_METADATA in definition:
            metadata_to_add = definition[AgentType._KEY_METADATA]
            if metadata_to_add:
                agent_type._metadata.update(metadata_to_add)
        else:
            log().debug(AgentType._NO_METADATA.format(name))

        return agent_type

    @staticmethod
    def _read_products(products: Any, agent_type: str) -> Dict[str, Any]:
        """Returns a dict obtained from given `products` defined for `agent_type`"""
        product_names = None
        if isinstance(products, dict):
            product_names = products
        elif isinstance(products, list):
            product_names = {key: None for key in products}
        else:
            log_error_and_raise(SchemaException(AgentType._ERR_PRODUCTS_UNKNOWN_STRUCTURE.format(agent_type, products)))

        if all([isinstance(item, str) for item in product_names.keys()]):
            return product_names
        else:
            log_error_and_raise(SchemaException(AgentType._ERR_PRODUCTS_NO_STRING.format(agent_type, product_names)))

    @property
    def name(self) -> str:
        """Returns the agent type name"""
        return self._name

    @property
    def products(self) -> dict:
        """Returns dict of products or an empty dict if no products are defined"""
        return self._products

    def get_product_names(self) -> List[str]:
        """Returns list of product names or an empty list if no products are defined"""
        return list(self._products.keys())

    @property
    def attributes(self) -> Dict[str, AttributeSpecs]:
        """Returns list of Attributes of this agent or an empty list if no attributes are defined"""
        return self._attributes

    @property
    def outputs(self) -> dict:
        """Returns list of outputs or an empty list if no outputs are defined"""
        return self._outputs

    @property
    def metadata(self) -> dict:
        """Returns list of metadata or an empty list if no metadata are defined"""
        return self._metadata



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Any, Dict, List

from fameio.source.scenario.exception import log_and_raise


class Attribute:
    """An Attribute of an agent in a scenario"""

    _VALUE_MISSING = "Value not specified for Attribute '{}' - leave out if default shall be used (if defined)."
    _LIST_EMPTY = "Attribute '{}' was assigned an empty list - please remove or fill empty assignments."
    _DICT_EMPTY = "Attribute '{}' was assigned an empty dictionary - please remove or fill empty assignments."
    _MIXED_DATA = "Attribute '{}' was assigned a list with mixed complex and simple entries - please fix."

    def __init__(self, name: str, definitions) -> None:
        """Parses an Attribute's definition"""
        self._full_name = name

        if definitions is None:
            log_and_raise(Attribute._VALUE_MISSING.format(name))

        if isinstance(definitions, dict):
            self._value = None
            self._nested_list = None
            self._nested = Attribute._build_attribute_dict(name, definitions)
        elif Attribute._is_list_of_dict(name, definitions):
            self._nested = None
            self._value = None
            self._nested_list = []
            for entry in definitions:
                self._nested_list.append(Attribute._build_attribute_dict(name, entry))
        else:
            self._nested = None
            self._nested_list = None
            self._value = definitions

    @staticmethod
    def _build_attribute_dict(name: str, definitions: Dict[str, Any]) -> Dict[str, "Attribute"]:
        """Returns a new dictionary containing Attributes generated from given `definitions`"""
        if not definitions:
            log_and_raise(Attribute._DICT_EMPTY.format(name))

        inner_elements = {}
        for nested_name, value in definitions.items():
            full_name = name + "." + nested_name
            inner_elements[nested_name] = Attribute(full_name, value)
        return inner_elements

    @staticmethod
    def _is_list_of_dict(name: str, definitions: Any) -> bool:
        """Returns True if given `definitions` is a list of dict"""
        if isinstance(definitions, list):
            if not definitions:
                log_and_raise(Attribute._LIST_EMPTY.format(name))

            all_dicts = no_dicts = True
            for item in definitions:
                if not isinstance(item, dict):
                    all_dicts = False
                else:
                    no_dicts = False
            if (not all_dicts) and (not no_dicts):
                log_and_raise(Attribute._MIXED_DATA.format(name))
            return all_dicts
        return False

    @property
    def generic_content(self) -> Any:
        """Returns the full content of the attribute (and its children) as a generic value"""
        if self.has_value:
            return self.value
        elif self.has_nested_list:
            result = []
            for attr_dict in self.nested_list:
                inner_elements = {}
                for name, attr in attr_dict.items():
                    inner_elements[name] = attr.generic_content
                result.append(inner_elements)
            return result
        elif self.has_nested:
            result = {}
            for name, attr in self.nested.items():
                result[name] = attr.generic_content
            return result
        else:
            log_and_raise(Attribute._VALUE_MISSING.format(self._full_name))

    @property
    def has_value(self) -> bool:
        """Returns True if Attribute has any value assigned"""
        return self._value is not None

    @property
    def value(self) -> Any:
        return self._value

    @property
    def has_nested(self) -> bool:
        """Returns True if nested Attributes are present"""
        return bool(self._nested)

    @property
    def nested(self) -> Dict[str, Attribute]:
        """Returns dictionary of all nested Attributes"""
        assert self.has_nested
        return self._nested

    def get_nested_by_name(self, key: str) -> Attribute:
        """Returns nested Attribute by specified name"""
        return self._nested[key]

    @property
    def has_nested_list(self) -> bool:
        """Returns True if list of nested items is present"""
        return bool(self._nested_list)

    @property
    def nested_list(self) -> List[Dict[str, "Attribute"]]:
        """Return list of all nested Attribute dictionaries"""
        assert self.has_nested_list
        return self._nested_list

    def __repr__(self) -> str:
        return self._full_name



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import glob
import os
from typing import List, Optional


class PathResolver:
    """Class responsible for locating files referenced in a scenario.

    Such files can be the ones referenced via the YAML `!include` extension, or simply the data files (time_series)
    referenced in attributes.

    This class provides a default behaviour that can easily be customized by the caller."""

    def resolve_yaml_imported_file_pattern(self, root_path: str, file_pattern: str) -> List[str]:
        """Returns a list of file paths matching the given `file_pattern` based on the configured resolver."""
        absolute_path = os.path.abspath(os.path.join(root_path, file_pattern))
        return glob.glob(absolute_path)

    def resolve_series_file_path(self, file_name: str) -> Optional[str]:
        """Returns the absolute file path for the given series (relative) file name, or None on failure."""
        if os.path.isabs(file_name):
            return file_name

        # try to locate in the current dir
        file_path = os.path.join(os.path.curdir, file_name)
        if os.path.exists(file_path):
            return file_path

        # not found
        return None



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import copy
from argparse import ArgumentParser, ArgumentTypeError, BooleanOptionalAction, Namespace
from enum import Enum
from pathlib import Path
from typing import Optional, Dict, Any, List, Union

from fameio.source.cli.options import MergingOptions, TimeOptions, ResolveOptions, Options
from fameio.source.logs import LogLevel

_ERR_NEGATIVE_INT = "Given value `{}` is not a non-negative int."

_OPTION_ARGUMENT_NAME: Dict[str, Union[Options, Dict]] = {
    "file": Options.FILE,
    "log": Options.LOG_LEVEL,
    "logfile": Options.LOG_FILE,
    "output": Options.OUTPUT,
    "agents": Options.AGENT_LIST,
    "single_export": Options.SINGLE_AGENT_EXPORT,
    "memory_saving": Options.MEMORY_SAVING,
    "time": Options.TIME,
    "input_recovery": Options.INPUT_RECOVERY,
    "complex_column": Options.RESOLVE_COMPLEX_FIELD,
    "time_merging": {
        "name": Options.TIME_MERGING,
        "inner_elements": {
            "focal_point": MergingOptions.FOCAL_POINT,
            "steps_before": MergingOptions.STEPS_BEFORE,
            "steps_after": MergingOptions.STEPS_AFTER,
        },
    },
}


def add_file_argument(parser: ArgumentParser, default: Optional[Path], help_text: str) -> None:
    """
    Adds 'file' argument to the provided `parser` with the provided `help_text`.
    If a default is not specified, the argument is required (optional otherwise)

    Args:
        parser: to add the argument to
        default: optional, if it is a valid Path, it is added as default and the argument becomes optional
        help_text: to be displayed
    """
    if default is not None and isinstance(default, (Path, str)):
        parser.add_argument("-f", "--file", type=Path, required=False, default=default, help=help_text)
    else:
        parser.add_argument("-f", "--file", type=Path, required=True, help=help_text)


def add_select_agents_argument(parser: ArgumentParser, default: List[str]) -> None:
    """Adds optional repeatable string argument 'agent' to given `parser`"""
    help_text = "Provide list of agents to extract (default=None)"
    parser.add_argument("-a", "--agents", nargs="*", type=str, default=default, help=help_text)


def add_logfile_argument(parser: ArgumentParser, default: Path) -> None:
    """Adds optional argument 'logfile' to given `parser`"""
    help_text = "provide logging file (default=None)"
    parser.add_argument("-lf", "--logfile", type=Path, default=default, help=help_text)


def add_output_argument(parser: ArgumentParser, default_value, help_text: str) -> None:
    """Adds optional argument 'output' to given `parser` using the given `help_text` and `default_value`"""
    parser.add_argument("-o", "--output", type=Path, default=default_value, help=help_text)


def add_log_level_argument(parser: ArgumentParser, default_value: str) -> None:
    """Adds optional argument 'log' to given `parser`"""
    help_text = "choose logging level (default: {})".format(default_value)
    parser.add_argument(
        "-l",
        "--log",
        default=default_value,
        choices=[level.name for level in LogLevel if level not in [LogLevel.PRINT, LogLevel.WARN]],
        type=str.upper,
        help=help_text,
    )


def add_single_export_argument(parser: ArgumentParser, default_value: bool) -> None:
    """Adds optional repeatable string argument 'agent' to given `parser`"""
    help_text = "Enable export of single agents (default=False)"
    parser.add_argument(
        "-se",
        "--single-export",
        default=default_value,
        action="store_true",
        help=help_text,
    )


def add_memory_saving_argument(parser: ArgumentParser, default_value: bool) -> None:
    """Adds optional bool argument to given `parser` to enable memory saving mode"""
    help_text = "Reduces memory usage profile at the cost of runtime (default=False)"
    parser.add_argument(
        "-m",
        "--memory-saving",
        default=default_value,
        action="store_true",
        help=help_text,
    )


def add_resolve_complex_argument(parser: ArgumentParser, default_value: Union[ResolveOptions, str]):
    """Instructs given `parser` how to deal with complex field outputs"""
    default_value = default_value if isinstance(default_value, ResolveOptions) else ResolveOptions[default_value]
    help_text = f"How to deal with complex index columns? (default={default_value})"
    parser.add_argument(
        "-cc",
        "--complex-column",
        type=ResolveOptions.instantiate,
        default=default_value,
        choices=ResolveOptions,
        help=help_text,
    )


def add_time_argument(parser: ArgumentParser, default_value: Union[TimeOptions, str]) -> None:
    """Adds optional argument to given `parser` to define conversion of TimeSteps"""
    default_value = default_value if isinstance(default_value, TimeOptions) else TimeOptions[default_value]
    help_text = "Apply conversion of time steps to given format (default=UTC)"
    parser.add_argument(
        "-t",
        "--time",
        type=TimeOptions.instantiate,
        default=default_value,
        choices=TimeOptions,
        help=help_text,
    )


def add_merge_time_parser(parser: ArgumentParser) -> None:
    """Adds subparser for merging of TimeSteps to given `parser`"""
    subparser = parser.add_subparsers(dest="time_merging", required=False, help="Optional merging of TimeSteps")
    group_parser = subparser.add_parser("merge-times")
    add_focal_point_argument(group_parser)
    add_steps_before_argument(group_parser)
    add_steps_after_argument(group_parser)


def add_focal_point_argument(parser: ArgumentParser) -> None:
    """Adds `focal-point` argument to given `parser`"""
    help_text = "TimeStep on which `steps_before` earlier and `steps_after` later TimeSteps are merged on"
    parser.add_argument("-fp", "--focal-point", required=True, type=int, help=help_text)


def add_steps_before_argument(parser: ArgumentParser) -> None:
    """Adds `steps-before` argument to given `parser`"""
    help_text = "Range of TimeSteps before the `focal-point` they get merged to"
    parser.add_argument("-sb", "--steps-before", required=True, type=_non_negative_int, help=help_text)


def _non_negative_int(value: Any) -> int:
    """
    Casts a given ´value` to int and checks it for non-negativity

    Args:
        value: to check and parse

    Returns:
        `value` parsed to int if it is a non-negative integer

    Raises:
        TypeError: if `value` is None
        ValueError: if `value` cannot be parsed to int
        argparse.ArgumentTypeError: if `value` is a negative int

    """
    value = int(value)
    if value < 0:
        raise ArgumentTypeError(_ERR_NEGATIVE_INT.format(value))
    return value


def add_steps_after_argument(parser: ArgumentParser) -> None:
    """Adds `steps-after` argument to given `parser`"""
    help_text = "Range of TimeSteps after the `focal-point` they get merged to"
    parser.add_argument("-sa", "--steps-after", required=True, type=_non_negative_int, help=help_text)


def add_inputs_recovery_argument(parser: ArgumentParser, default: bool) -> None:
    """Adds optional bool argument to given `parser` to recover inputs"""
    help_text = "If --(no-)input-recovery is specified, (no) inputs will be recovered"
    parser.add_argument(
        "--input-recovery",
        action=BooleanOptionalAction,
        default=default,
        help=help_text,
    )


def update_default_config(config: Optional[dict], default: dict) -> dict:
    """Returns `default` config with updated fields received from `config`"""
    result = copy.deepcopy(default)
    if config:
        for name, option in config.items():
            result[name] = option
    return result


def map_namespace_to_options_dict(parsed: Namespace) -> Dict[Options, Any]:
    """
    Maps given parsing results to their corresponding configuration option

    Args:
        parsed: result of a parsing

    Returns:
        Map of each parsed argument to their configuration option
    """
    return _map_namespace_to_options(parsed, _OPTION_ARGUMENT_NAME)


def _map_namespace_to_options(parsed: Namespace, names_to_options: Dict[str, Enum]) -> Dict[Options, Any]:
    """
    Maps given parsing results to their corresponding configuration option; elements that cannot be mapped are ignored.
    If a configuration option has inner elements, these well be also read and added as inner dictionary.

    Args:
        parsed: result of a parsing
        names_to_options: dict to search for configuration option specifications

    Returns:
         Map parsed arguments to their configuration option if they exist in the given `names_to_options` dict
    """
    config = {}
    for name, value in vars(parsed).items():
        option = names_to_options.get(name, None)
        if option:
            if isinstance(option, dict):
                inner_element_map = option["inner_elements"]
                option = option["name"]
                value = _map_namespace_to_options(parsed, inner_element_map)
            config[option] = value
    return config



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as pylog
from enum import Enum
from pathlib import Path
from typing import NoReturn, Optional, List


class LogLevel(Enum):
    """Levels for Logging"""

    PRINT = 100
    CRITICAL = pylog.CRITICAL
    ERROR = pylog.ERROR
    WARN = pylog.WARNING
    WARNING = pylog.WARNING
    INFO = pylog.INFO
    DEBUG = pylog.DEBUG


_loggers: List[pylog.Logger] = []
_handlers: List[pylog.Handler] = []

_FORMAT_NORMAL = "%(asctime)s — %(levelname)s — %(message)s"  # noqa
_FORMAT_DETAILLED = "%(asctime)s.%(msecs)03d — %(levelname)s — %(module)s:%(funcName)s:%(lineno)d — %(message)s"  # noqa
_TIME_FORMAT = "%H:%M:%S"

_INFO_UPDATING_LOG_LEVEL = "Updating fameio log level to: {}"
_WARN_NOT_INITIALIZED = "Logger for fameio not initialised: using default log level `WARNING`"

LOGGER_NAME = "fameio"
DEFAULT_LOG_LEVEL = LogLevel.WARNING


def log() -> pylog.Logger:
    """Returns already set up FAME-Io's logger or - if not set up - a new logger with `WARNING`"""
    if not _loggers:
        fameio_logger(DEFAULT_LOG_LEVEL.name)
        pylog.warning(_WARN_NOT_INITIALIZED)
    return _loggers[0]


def log_and_raise_critical(message: str) -> NoReturn:
    """Raises a critical error and logs with given `error_message`"""
    log().critical(message)
    raise Exception(message)


def log_error_and_raise(exception: Exception) -> NoReturn:
    """Raises the specified `exception` and logs an error with the same `message`"""
    log().error(str(exception))
    raise exception


def fameio_logger(log_level_name: str, file_name: Optional[Path] = None) -> None:
    """
    Ensures a logger for fameio is present and uses the specified options

    Args:
        log_level_name: one of Python's official logging level names, e.g. "INFO"
        file_name: if present, logs are also written to the specified file path
    """
    log_level = LogLevel[log_level_name.upper()]
    logger = _get_logger(log_level)

    formatter = _get_formatter(log_level)
    _add_handler(logger, pylog.StreamHandler(), formatter)
    if file_name:
        _add_handler(logger, pylog.FileHandler(file_name, mode="w"), formatter)

    if _loggers:
        pylog.info(_INFO_UPDATING_LOG_LEVEL.format(log_level_name))
        _loggers[0] = logger
    else:
        _loggers.append(logger)


def _get_logger(level: LogLevel) -> pylog.Logger:
    """
    Returns fameio logger with given log level without any handler and, not propagating to parent

    Args:
        level: integer representing the log level

    Returns:
        logger for fameio with specified level
    """
    logger = pylog.getLogger(LOGGER_NAME)
    logger.setLevel(level.value)
    logger.propagate = False
    for handler in _handlers:
        logger.removeHandler(handler)
    _handlers.clear()
    return logger


def _get_formatter(level: LogLevel) -> pylog.Formatter:
    """
    Returns a log formatter depending on the given log `level`
    Args:
        level: this log level determines how detailed the logger's output is
    Returns:
        new log formatter
    """
    return pylog.Formatter(_FORMAT_DETAILLED if level is LogLevel.DEBUG else _FORMAT_NORMAL, _TIME_FORMAT)


def _add_handler(logger: pylog.Logger, handler: pylog.Handler, formatter: pylog.Formatter) -> None:
    """Adds given `handler` using the specified `formatter` to given `logger` and `_handlers` list"""
    handler.setFormatter(formatter)
    _handlers.append(handler)
    logger.addHandler(handler)



# SPDX-FileCopyrightText: 2024 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from typing import Dict

import yaml

from fameio.source.logs import log

ERR_WRITE_EXCEPTION = "Failed to save dictionary to YAML file `{}`"
INFO_DESTINATION = "Saving scenario to file at {}"


def data_to_yaml_file(data: Dict, file_path: Path) -> None:
    """
    Save the given data to a YAML file at given path

    Args:
        data: to be saved to yaml file
        file_path: at which the file will be created
    """
    log().info(INFO_DESTINATION.format(file_path))
    try:
        with open(file_path, "w") as f:
            yaml.dump(data, f, sort_keys=False)
    except Exception as e:
        raise RuntimeError(ERR_WRITE_EXCEPTION.format(file_path)) from e



#!/usr/bin/env python
import sys
from pathlib import Path

from fameio.source.cli.make_config import handle_args, CLI_DEFAULTS as DEFAULT_CONFIG
from fameio.source.cli.options import Options
from fameio.source.cli.parser import update_default_config
from fameio.source.loader import load_yaml, check_for_yaml_file_type
from fameio.source.logs import fameio_logger, log
from fameio.source.scenario import Scenario
from fameio.source.validator import SchemaValidator
from fameio.source.writer import ProtoWriter


def run(config: dict = None) -> None:
    """Executes the main workflow for the building of a FAME configuration file"""
    config = update_default_config(config, DEFAULT_CONFIG)
    fameio_logger(log_level_name=config[Options.LOG_LEVEL], file_name=config[Options.LOG_FILE])

    file = config[Options.FILE]
    check_for_yaml_file_type(Path(file))
    scenario = Scenario.from_dict(load_yaml(Path(file)))
    SchemaValidator.check_agents_have_contracts(scenario)

    timeseries_manager = SchemaValidator.validate_scenario_and_timeseries(scenario)
    writer = ProtoWriter(config[Options.OUTPUT], timeseries_manager)
    writer.write_validated_scenario(scenario)

    log().info("Configuration completed.")


if __name__ == "__main__":
    run_config = handle_args(sys.argv[1:])
    run(run_config)



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import datetime as dt
import math
import re
from enum import Enum, auto
from typing import Union

from fameio.source.logs import log_error_and_raise

START_IN_REAL_TIME = "2000-01-01_00:00:00"
DATE_FORMAT = "%Y-%m-%d_%H:%M:%S"
DATE_REGEX = re.compile("[0-9]{4}-[0-9]{2}-[0-9]{2}_[0-9]{2}:[0-9]{2}:[0-9]{2}")
FAME_FIRST_DATETIME = dt.datetime.strptime(START_IN_REAL_TIME, DATE_FORMAT)


class ConversionException(Exception):
    """Indicates that something went wrong during time stamp conversion"""

    pass


class TimeUnit(Enum):
    """Time units defined in FAME"""

    SECONDS = auto()
    MINUTES = auto()
    HOURS = auto()
    DAYS = auto()
    WEEKS = auto()
    MONTHS = auto()
    YEARS = auto()


class Constants:
    """Time steps in FAME simulations associated with corresponding TimeUnits"""

    FIRST_YEAR = 2000
    STEPS_PER_SECOND = 1
    SECONDS_PER_MINUTE = 60
    MINUTES_PER_HOUR = 60
    HOURS_PER_DAY = 24
    DAYS_PER_YEAR = 365
    STEPS_PER_MINUTE = STEPS_PER_SECOND * SECONDS_PER_MINUTE
    STEPS_PER_HOUR = STEPS_PER_MINUTE * MINUTES_PER_HOUR
    STEPS_PER_DAY = STEPS_PER_HOUR * HOURS_PER_DAY
    STEPS_PER_YEAR = STEPS_PER_DAY * DAYS_PER_YEAR
    STEPS_PER_WEEK = STEPS_PER_DAY * 7
    STEPS_PER_MONTH = STEPS_PER_YEAR / 12

    steps_per_unit = {
        TimeUnit.SECONDS: STEPS_PER_SECOND,
        TimeUnit.MINUTES: STEPS_PER_MINUTE,
        TimeUnit.HOURS: STEPS_PER_HOUR,
        TimeUnit.DAYS: STEPS_PER_DAY,
        TimeUnit.WEEKS: STEPS_PER_WEEK,
        TimeUnit.MONTHS: STEPS_PER_MONTH,
        TimeUnit.YEARS: STEPS_PER_YEAR,
    }


class FameTime:
    """Handles conversion of TimeSteps and TimeDurations into TimeStamps and vice versa"""

    _TIME_UNIT_UNKNOWN = "TimeUnit conversion of '{}' not implemented."
    _FORMAT_INVALID = "'{}' is not recognised as time stamp string - check its format."
    _INVALID_TIMESTAMP = "Cannot convert time stamp string '{}' - check its format."
    _INVALID_TOO_LARGE = "Cannot convert time stamp string '{}' - last day of leap year is Dec 30th!"
    _NO_TIMESTAMP = "Time value expected, but '{}' is neither a time stamp string nor an integer."
    _INVALID_DATE_FORMAT = "Received invalid date format '{}'."

    @staticmethod
    def convert_datetime_to_fame_time_step(datetime_string: str) -> int:
        """Converts real Datetime string to FAME time step"""
        if not FameTime.is_datetime(datetime_string):
            log_error_and_raise(ConversionException(FameTime._FORMAT_INVALID.format(datetime_string)))
        datetime = FameTime._convert_to_datetime(datetime_string)
        years_since_start_time = datetime.year - FAME_FIRST_DATETIME.year
        beginning_of_year = dt.datetime(year=datetime.year, month=1, day=1, hour=0, minute=0, second=0)
        seconds_since_beginning_of_year = int((datetime - beginning_of_year).total_seconds())
        steps_since_beginning_of_year = seconds_since_beginning_of_year * Constants.STEPS_PER_SECOND
        if steps_since_beginning_of_year > Constants.STEPS_PER_YEAR:
            log_error_and_raise(ConversionException(FameTime._INVALID_TOO_LARGE.format(datetime_string)))
        year_offset = years_since_start_time * Constants.STEPS_PER_YEAR
        return year_offset + steps_since_beginning_of_year

    @staticmethod
    def _convert_to_datetime(datetime_string: str) -> dt.datetime:
        """Converts given `datetime_string` to real-world datetime"""
        try:
            return dt.datetime.strptime(datetime_string, DATE_FORMAT)
        except ValueError:
            log_error_and_raise(ConversionException(FameTime._INVALID_TIMESTAMP.format(datetime_string)))

    @staticmethod
    def convert_fame_time_step_to_datetime(fame_time_steps: int, date_format: str = DATE_FORMAT) -> str:
        """
        Converts given `fame_time_steps` to corresponding real-world datetime string in `date_format`,
        raises ConversionException if invalid `date_format` received.
        """
        years_since_start_time = math.floor(fame_time_steps / Constants.STEPS_PER_YEAR)
        current_year = years_since_start_time + Constants.FIRST_YEAR
        beginning_of_year = dt.datetime(year=current_year, month=1, day=1, hour=0, minute=0, second=0)
        steps_in_current_year = fame_time_steps - years_since_start_time * Constants.STEPS_PER_YEAR
        seconds_in_current_year = steps_in_current_year / Constants.STEPS_PER_SECOND
        datetime = beginning_of_year + dt.timedelta(seconds=seconds_in_current_year)
        try:
            return datetime.strftime(date_format)
        except ValueError:
            log_error_and_raise(ConversionException(FameTime._INVALID_DATE_FORMAT.format(date_format)))

    @staticmethod
    def convert_time_span_to_fame_time_steps(value: int, unit: TimeUnit) -> int:
        """Converts value of `TimeUnit.UNIT` to fame time steps"""
        steps = Constants.steps_per_unit.get(unit)
        if steps:
            return steps * value
        else:
            log_error_and_raise(ConversionException(FameTime._TIME_UNIT_UNKNOWN.format(unit)))

    @staticmethod
    def is_datetime(string: str) -> bool:
        """Returns `True` if given `string` matches Datetime string format and can be converted to FAME time step"""
        if isinstance(string, str):
            return DATE_REGEX.fullmatch(string.strip()) is not None
        return False

    @staticmethod
    def is_fame_time_compatible(value: Union[int, str]) -> bool:
        """Returns `True` if given int or string `value` can be converted to a FAME time step"""
        if isinstance(value, int):
            return True
        elif isinstance(value, str):
            return FameTime.is_datetime(value) or FameTime._is_integer(value)
        return False

    @staticmethod
    def _is_integer(string: str) -> bool:
        """Returns `True` if given string can be interpreted as integer"""
        try:
            int(string)
        except ValueError:
            return False
        else:
            return True

    @staticmethod
    def convert_string_if_is_datetime(value: Union[int, str]) -> int:
        """
        Returns FAME time steps If given `value` is a valid FAME datetime string it is converted to FAME time steps;
        Or, if given `value` is an integer, it is returned without modification.
        Raises an Exception if given `value` is neither a valid FAME datetime string nor an integer value
        """
        if FameTime.is_datetime(value):
            return int(FameTime.convert_datetime_to_fame_time_step(value))
        else:
            try:
                return int(value)
            except ValueError:
                log_error_and_raise(ConversionException(FameTime._NO_TIMESTAMP.format(value)))



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from typing import NoReturn, Any, Union

from fameio.source.logs import log_error_and_raise, log

_DEFAULT_USED = "Using default value '{}' for missing key '{}'"


def log_and_raise(message: str) -> NoReturn:
    """Raises ScenarioException with given `message`"""
    log_error_and_raise(ScenarioException(message))


def get_or_raise(dictionary: dict, key: str, message: str) -> Union[Any, NoReturn]:
    """Returns value associated with `key` in given `dictionary`, or raises ScenarioException if key is missing"""
    if key not in dictionary or dictionary[key] is None:
        log_error_and_raise(ScenarioException(message.format(key)))
    else:
        return dictionary[key]


def assert_or_raise(assertion: bool, msg: str) -> None:
    """Raises new ScenarioException with given `msg` if `assertion` is False"""
    if not assertion:
        log_error_and_raise(ScenarioException(msg))


def get_or_default(dictionary: dict, key: str, default_value) -> Any:
    """Returns value associated with `key` in given `dictionary`, or the given `default_value` if key is missing"""
    if key in dictionary and dictionary[key] is not None:
        return dictionary[key]
    else:
        log().debug(_DEFAULT_USED.format(default_value, key))
        return default_value


class ScenarioException(Exception):
    """Indicates an error while parsing a scenario or one of its components"""

    pass



# SPDX-FileCopyrightText: 2023 German Aerospace Center <fame@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import math
import os
from enum import Enum, auto
from pathlib import Path
from typing import Dict, Union, Tuple, Any, List

import pandas as pd
from fameprotobuf.InputFile_pb2 import InputData
from google.protobuf.internal.wire_format import INT64_MIN, INT64_MAX

from fameio.source import PathResolver
from fameio.source.logs import log_error_and_raise, log
from fameio.source.time import ConversionException, FameTime
from fameio.source.tools import clean_up_file_name


class TimeSeriesException(Exception):
    """Indicates that an error occurred during management of time series"""

    pass


class Entry(Enum):
    ID = auto()
    NAME = auto()
    DATA = auto()


class TimeSeriesManager:
    """Manages matching of files to time series ids and their protobuf representation"""

    _TIMESERIES_RECONSTRUCTION_PATH = "./timeseries/"
    _CONSTANT_IDENTIFIER = "Constant value: {}"
    _KEY_ROW_TIME = "timeStep"
    _KEY_ROW_VALUE = "value"

    _ERR_FILE_NOT_FOUND = "Cannot find Timeseries file '{}'"
    _ERR_CORRUPT_TIME_SERIES_KEY = "TimeSeries file '{}' corrupt: At least one entry in first column isn't a timestamp."
    _ERR_CORRUPT_TIME_SERIES_VALUE = "TimeSeries file '{}' corrupt: At least one entry in value column isn't numeric."
    _ERR_NON_NUMERIC = "Values in TimeSeries must be numeric but was: '{}'"
    _ERR_NAN_VALUE = "Values in TimeSeries must not be missing or NaN."
    _ERR_UNREGISTERED_SERIES = "No timeseries registered with identifier '{}' - was the Scenario validated?"
    _WARN_NO_DATA = "No timeseries stored in timeseries manager. Double check if you expected timeseries."

    def __init__(self, path_resolver: PathResolver = PathResolver()) -> None:
        self._path_resolver = path_resolver
        self._id_count = -1
        self._series_by_id: Dict[Union[str, int, float], Dict[Entry, Any]] = {}

    def register_and_validate(self, identifier: Union[str, int, float]) -> None:
        """
        Registers given timeseries `identifier` and validates associated timeseries

        Args:
            identifier: to be registered - either a single numeric value or a string pointing to a timeseries file

        Raises:
            TimeSeriesException: if file was not found, ill-formatted, or value was invalid
        """
        if not self._time_series_is_registered(identifier):
            self._register_time_series(identifier)

    def _time_series_is_registered(self, identifier: Union[str, int, float]) -> bool:
        """Returns True if the value was already registered"""
        return identifier in self._series_by_id.keys()

    def _register_time_series(self, identifier: Union[str, int, float]) -> None:
        """Assigns an id to the given `identifier` and loads the time series into a dataframe"""
        self._id_count += 1
        name, series = self._get_name_and_dataframe(identifier)
        self._series_by_id[identifier] = {Entry.ID: self._id_count, Entry.NAME: name, Entry.DATA: series}

    def _get_name_and_dataframe(self, identifier: Union[str, int, float]) -> Tuple[str, pd.DataFrame]:
        """Returns name and DataFrame containing the series obtained from the given `identifier`"""
        if isinstance(identifier, str):
            series_path = self._path_resolver.resolve_series_file_path(Path(identifier).as_posix())
            if series_path and os.path.exists(series_path):
                data = pd.read_csv(series_path, sep=";", header=None, comment="#")
                try:
                    return identifier, self._check_and_convert_series(data)
                except TypeError as e:
                    log_error_and_raise(TimeSeriesException(self._ERR_CORRUPT_TIME_SERIES_VALUE.format(identifier), e))
                except ConversionException:
                    log_error_and_raise(TimeSeriesException(self._ERR_CORRUPT_TIME_SERIES_KEY.format(identifier)))
            else:
                log_error_and_raise(TimeSeriesException(self._ERR_FILE_NOT_FOUND.format(identifier)))
        else:
            return self._create_timeseries_from_value(identifier)

    def _check_and_convert_series(self, data: pd.DataFrame) -> pd.DataFrame:
        """Ensures validity of time series and convert to required format for writing to disk"""
        data = data.apply(
            lambda r: [FameTime.convert_string_if_is_datetime(r[0]), self._assert_valid(r[1])],
            axis=1,
            result_type="expand",
        )
        return data

    @staticmethod
    def _assert_valid(value: Any) -> float:
        """Returns the given `value` if it is a numeric value other than NaN"""
        try:
            value = float(value)
        except ValueError:
            log_error_and_raise(TypeError(TimeSeriesManager._ERR_NON_NUMERIC.format(value)))
        if math.isnan(value):
            log_error_and_raise(TypeError(TimeSeriesManager._ERR_NAN_VALUE))
        return value

    @staticmethod
    def _create_timeseries_from_value(value: Union[int, float]) -> Tuple[str, pd.DataFrame]:
        """Returns name and dataframe for a new static timeseries created from the given `value`"""
        if math.isnan(value):
            log_error_and_raise(TimeSeriesException(TimeSeriesManager._ERR_NAN_VALUE))
        data = pd.DataFrame({0: [INT64_MIN, INT64_MAX], 1: [value, value]})
        return TimeSeriesManager._CONSTANT_IDENTIFIER.format(value), data

    def get_series_id_by_identifier(self, identifier: Union[str, int, float]) -> int:
        """
        Returns id for a previously stored time series by given `identifier`

        Args:
            identifier: to get the unique ID for

        Returns:
            unique ID for the given identifier

        Raises:
            TimeSeriesException: if identifier was not yet registered
        """
        if not self._time_series_is_registered(identifier):
            log_error_and_raise(TimeSeriesException(self._ERR_UNREGISTERED_SERIES.format(identifier)))
        return self._series_by_id.get(identifier)[Entry.ID]

    def get_all_series(self) -> List[Tuple[int, str, pd.DataFrame]]:
        """Returns iterator over id, name and dataframe of all stored series"""
        if len(self._series_by_id) == 0:
            log().warning(self._WARN_NO_DATA)
        return [(v[Entry.ID], v[Entry.NAME], v[Entry.DATA]) for v in self._series_by_id.values()]

    def reconstruct_time_series(self, timeseries: List[InputData.TimeSeriesDao]) -> None:
        """Reconstructs and stores time series from given list of `timeseries_dao`"""
        for one_series in timeseries:
            self._id_count += 1
            reconstructed = {Entry.ID: one_series.seriesId}
            if len(one_series.row) == 1:
                reconstructed[Entry.NAME] = one_series.row[0].value
                reconstructed[Entry.DATA] = None
            else:
                reconstructed[Entry.NAME] = self._get_cleaned_file_name(one_series.seriesName)
                reconstructed[Entry.DATA] = pd.DataFrame(
                    [{self._KEY_ROW_TIME: item.timeStep, self._KEY_ROW_VALUE: item.value} for item in one_series.row]
                )
            self._series_by_id[one_series.seriesId] = reconstructed

    def _get_cleaned_file_name(self, timeseries_name: str):
        if timeseries_name.lower().endswith(".csv"):
            filename = Path(timeseries_name).name
        else:
            filename = clean_up_file_name(timeseries_name) + ".csv"
        return str(Path(self._TIMESERIES_RECONSTRUCTION_PATH, filename))

    def get_reconstructed_series_by_id(self, series_id: int) -> str:
        """Return name or path for given `series_id` if series these are identified by their number.
        Use this only if series were added via `reconstruct_time_series`"""
        if series_id < 0 or series_id > self._id_count:
            log_error_and_raise(TimeSeriesException(self._ERR_UNREGISTERED_SERIES.format(series_id)))
        return self._series_by_id[series_id][Entry.NAME]
